import os
import platform
import sys
from itertools import chain


IS_WINDOWS = (platform.system() == 'Windows')
IS_DARWIN = (platform.system() == 'Darwin')
IS_LINUX = (platform.system() == 'Linux')


IS_CONDA = 'conda' in sys.version or 'Continuum' in sys.version or any([x.startswith('CONDA') for x in os.environ])
CONDA_DIR = os.path.join(os.path.dirname(sys.executable), '..')


def check_env_flag(name, default=''):
    return os.getenv(name, default).upper() in ['ON', '1', 'YES', 'TRUE', 'Y']


def check_negative_env_flag(name, default=''):
    return os.getenv(name, default).upper() in ['OFF', '0', 'NO', 'FALSE', 'N']


def gather_paths(env_vars):
    return list(chain(*(os.getenv(v, '').split(os.pathsep) for v in env_vars)))


def lib_paths_from_base(base_path):
    return [os.path.join(base_path, s) for s in ['lib/x64', 'lib', 'lib64']]


"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/7/23 9:59
@Function: visualization.py
@Contact: cuijinghao@tgqs.net
"""
from tgqSim.circuit.common_gates import (BASE_SINGLE_GATE,
                                         BASE_DOUBLE_GATE,
                                         BASE_TRIPLE_GATE,
                                         CONTROLLED_GATE,
                                         ROTATION_GATE)

# contain all the gates available, grant each gate a symbol to represent
# for controlled gates, one node is 'C', the other node is its symbol string, e.g. 'NOT'
GATE_SYMBOL_DICT = {
    'x': 'X', 'y': 'Y', 'z': 'Z',
    'rzz': 'Z', 'ryy': 'Y', 'rxx': 'X',

                    'cp': 'CPhase',
                    'cx': 'NOT', 'cz': 'Z', 'ccx': 'NOT',
                    's': 'S', 'sdg': 'Sdg', 't': 'T', 'tdg': 'Tdg',
                    'h': 'H', 'u3': 'U3',
                    'rx': 'Rx', 'ry': 'Ry', 'rz': 'Rz',
                    'syc': 'SYC',
                    'swap': 'SWAP', 'cswap': 'SWAP', 'iswap': 'ISWAP'
}


def get_gate_symbol(gatename: str):
    pass


def get_circuit_deep(qubitNum: int, gate_list: list) -> int:
    """
    获取整个线路图最大的深度

    Args:
        qubitNum (int): 线路图中比特数
        gate_list (list): 量子门序列

    Returns:
        int: 量子线路图的深度
    """
    # 初始化，每一个比特位起始值为0
    count = [0 for _ in range(qubitNum)]
    for pos, _, __ in gate_list:
        if isinstance(pos, int):
            # 单比特：若是比特位使用int类型表示，则直接在该比特位计数加一
            count[pos] += 1
        else:
            length = len(pos)
            if 1 == length:
                # 单比特：利用list表示作用比特位时，计数也是直接加一
                count[pos[0]] += 1
            else:
                # 多比特，主要是针对双比特门和三比特门
                # 拿到多比特门跨度所有比特的门计数中最大的门计数
                maxVal = max([count[i] for i in pos])
                # 更新多比特门跨度的所有比特对应的门计数（保证是左对齐）
                for i in range(min(pos), max(pos) + 1):
                    count[i] = maxVal + 1

    return max(count)


def update_index_list(indexList: list, pos: list) -> list:
    """
    获取所有比特位对应的指针位置，保真整个线路图是左对齐

    Args:
        indexList (list): 所有比特中，当前指针所指的位置
        pos (list): 比特门作用的比特位置

    Returns:
        list: 更新后指针的位置
    """
    if 1 == len(pos):
        # 单比特情形
        newCol = indexList[2 * pos[0]] + 1
    else:
        # 多比特情形
        # 拿到多比特门跨度所有比特对应指针最大的位置
        newCol = max([indexList[i] for i in range(2 * min(pos), 2 * max(pos) + 1)]) + 1
    # update indexList
    # 开始更新多比特门跨度所有比特对应指针的值
    for i in range(2 * min(pos), 2 * max(pos) + 1):
        indexList[i] = newCol
    return indexList


def draw_on_quite_gate(diagInfo: list, pos: list, newColIndex: int, diagEle: tuple)->list:
    """
    根据给出来的比特门作用的比特位置、列指针信息以及需要绘画的内容，在对应的比特位置上画出量子门
    这里面仅仅门作用位置的信息，不会更新连线

    Args:
        diagInfo (list): 画板信息，线路形状都会在这个上面呈现
        pos (list): 量子门作用的位置信息，确定行，长度与diagEle保持一致，在定义门确定好
        newColIndex (int): 绘制量子门对应的列索引，确定列
        diagEle (tuple): 绘制图形样子，长度与pos保持一致，由门控制，在定义门确定好

    Returns:
        list: 更新后的画板
    """
    for i, qIndex in enumerate(pos):
        offset = len(diagEle[i])
        # print(pos, diagEle[i])
        diagInfo[2 * qIndex] = diagInfo[2 * qIndex][:newColIndex] + diagEle[i] + diagInfo[2 * qIndex][newColIndex + offset:]
    return diagInfo


def draw_vertical_line(diagInfo: list, pos: list, newColIndex: int)->list:
    """
    依据门信息，画出对应的连线

    Args:
        diagInfo (list): 画板信息，线路形状都会在这个上面呈现
        pos (list): 量子门作用的位置信息，确定行
        newColIndex (int):  绘制量子门对应的列索引，确定列

    Returns:
        list: 更新后的画板
    """
    for i in range(2 * min(pos), 2 * max(pos) + 1):
        if i % 2 == 0:
            qIndex = i // 2
            if qIndex in pos:
                # 这个位置是门作用的位置，不能覆盖
                continue
            else:
                # 若是当前位置比特位，则利用垂直线替代┼
                diagInfo[i] = diagInfo[i][:newColIndex] + "┼" + diagInfo[i][newColIndex + 1:]
        else:
            # 若是当前位置是两个相邻比特之间的空隙，直接用│替代
            diagInfo[i] = diagInfo[i][:newColIndex] + "│" + diagInfo[i][newColIndex + 1:]
    return diagInfo


def draw_gate(diagInfo: list, pos: list, newColIndex: int, diagEle: tuple)->list:
    """
    在线路图中绘制量子门
    绘制方法：
    首先在量子门作用位置进行绘制，即先绘制从控制比特和目标比特的图形
    其次再绘制之间的连线

    Args:
        diagInfo (list): 画板信息，线路形状都会在这个上面呈现
        pos (list): 量子门作用的位置信息，确定行，长度与diagEle保持一致，在定义门确定好
        newColIndex (int): 绘制量子门对应的列索引，确定列
        diagEle (tuple): 绘制图形样子，长度与pos保持一致，由门控制，在定义门确定好

    Returns:
        list: 更新后的画板
    """
    diagInfo = draw_on_quite_gate(diagInfo=diagInfo, pos=pos, newColIndex=newColIndex, diagEle=diagEle)
    diagInfo = draw_vertical_line(diagInfo=diagInfo, pos=pos, newColIndex=newColIndex)
    return diagInfo


def to_text_diag(gates: list, width: int) -> str:
    """
    绘制量子线路图

    Args:
        circuit (tgqSim.QuantumCircuit): 量子线路图信息

    Returns:
        str: 可供打印的线路图信息
    """
    # use displayname_list element of quantumcircuit
    gateList = gates
    qubitNum = width
    indexList = [0 for _ in range(2 * qubitNum - 1)]
    circuit_deep = get_circuit_deep(qubitNum, gateList)
    # print(circuit_deep)
    # 奇数行绘制横线，偶数行绘制空格
    base_line, base_space = "───────────", "           "
    # 初始化画板样式
    diagInfo = [f"{i // 2}: " + base_line * circuit_deep if i % 2 == 0 else base_space * circuit_deep for i in range(2 * qubitNum - 1)]
    # e.g. gate_pos: Union[int, str], display_name: tuple, *gate_info
    for pos, display_name, gate in gateList:
        if isinstance(pos, int):
            pos = [pos]
        newColIndex = max([indexList[ele] * len(base_space) + 4 for ele in range(2 * min(pos), 2 * max(pos) + 1)])
        indexList = update_index_list(indexList=indexList, pos=pos)
        # todo: use mapping instead of element of gatelist, gatelist will not contain symbol
        diagInfo = draw_gate(diagInfo=diagInfo, pos=pos, newColIndex=newColIndex, diagEle=display_name)
    return "\n".join(diagInfo)

# class Visualizer:
#     def __init__(self):
#         self.diagram_str = ''
#         self.diagram = ''


def to_text_diagram(gates: list, width: int):
    # [(0, ('h',)), ([0, 1], ('cx',)), (0, ('rz', 0.708)), (0, ('rz', 0.0)), (0, ('rz', -0.5)),
    #  (1, ('u3', 0.708, 0.708, 0.708)), ([0, 1], ('measure',))]
    max_gate_length = max(len(GATE_SYMBOL_DICT[gate[1][0]]) for gate in gates) if gates else 1
    # colum_width refers to gate width, while width refers to the number of qubits
    column_width = max_gate_length + 2
    diagram_str = ['-' * (len(gates) * column_width + 1) for _ in range(width)]
    for col, (qubits, gate) in enumerate(gates):
        print(col, gate, qubits)
        gate = gate[0].lower()
        if isinstance(qubits, int):
            qubits = [qubits]
        # todo: not all two bit gates are control gates, need to consider gates like RXX, ISWAP etc
        # single gate
        if gate in BASE_SINGLE_GATE:
            for q in range(width):
                if q in qubits:
                    diagram_str[q] = (diagram_str[q][:col * column_width + 1]
                                      + f'{GATE_SYMBOL_DICT[gate]}'.center(column_width)
                                      + diagram_str[q][col * column_width + column_width + 1:])
                else:
                    diagram_str[q] = diagram_str[q][:col * column_width + 1] + '|'.center(column_width) + \
                                     diagram_str[q][
                                     col * column_width + column_width + 1:]
        # double gate
        elif gate in BASE_DOUBLE_GATE:
            # for controlled gates
            if gate in CONTROLLED_GATE:
                # by default, first returned result is control bit
                control, target = qubits
                for q in range(width):
                    if q == control:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + f'@'.center(column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
                    elif q == target:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + f'{GATE_SYMBOL_DICT[gate]}'.center(
                            column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
                    elif control < q < target or target < q < control:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + '|'.center(column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
                    else:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + ' '.center(column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
            # for non controlled gates
            else:
                bit0, bit1 = qubits
                for q in range(width):
                    if q == bit0:
                        diagram_str[q] = (diagram_str[q][:col * column_width + 1] +
                                          f'{GATE_SYMBOL_DICT[gate]}'.center(column_width) +
                                         diagram_str[q][col * column_width + column_width + 1:])
                    elif q == bit1:
                        diagram_str[q] = (diagram_str[q][:col * column_width + 1] +
                                          f'{GATE_SYMBOL_DICT[gate]}'.center(column_width) +
                                         diagram_str[q][col * column_width + column_width + 1:])
                    elif bit0 < q < bit1 or bit1 < q < bit0:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + '|'.center(column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
                    else:
                        diagram_str[q] = diagram_str[q][:col * column_width + 1] + ' '.center(column_width) + \
                                         diagram_str[q][
                                         col * column_width + column_width + 1:]
        elif gate in BASE_TRIPLE_GATE:
            pass
    diagram = '\n'.join(diagram_str)
    return diagram







import torch
import math
import torch_npu
# from tgqSim import QuantumCircuit
# from numba import njit, prange

# 定义复数乘法函数，接受四个实数作为输入，分别代表两个复数的实部和虚部
def complex_mul(a_real, a_imag, b_real, b_imag):
    # print()
    # print("a_real:", a_real)
    # print("a_imag:", a_imag)
    # print("b_real:", b_real)
    # print("b_imag:", b_imag)
    result = [a_real * b_real - a_imag * b_imag, a_real * b_imag + a_imag * b_real]
    # print("complex_mul result:", result)
    return result

def complex_exp(real, imag):
    exp_real = torch.exp(real)
    return [exp_real * torch.cos(imag), exp_real * torch.sin(imag)]

def ActOn_State(psi: list, num_qubits, Gate_type, Gate_pos, *Angles):
    psi_real = torch.tensor(psi, dtype=torch.float32)
    psi_imag = torch.tensor([0]*len(psi), dtype=torch.float32)
    if torch.npu.is_available():
        device = torch.device("npu")
    else:
        device = torch.device("cpu")
    psi_real = psi_real.to(device)
    psi_imag = psi_imag.to(device)
    
    j1, j0 = max(Gate_pos), min(Gate_pos)
    i1_plus = 2 ** j0
    i2_plus = 2 ** j1
    i3_plus = 2 ** j0 + 2 ** j1
    delta2 = 2 ** (j1 + 1)
    delta1 = 2 ** (j0 + 1)
    max2 = 2 ** num_qubits - delta2
    max1 = 2 ** j1 - delta1
    max0 = 2 ** j0 - 1

    for k in range(0, max2 + 1, delta2):
        for l in range(0, max1 + 1, delta1):
            for m in range(0, max0 + 1):
                i0 = m | l | k
                i1 = i0 | i1_plus
                i2 = i0 | i2_plus
                i3 = i0 | i3_plus
                if (i0 != i1) and (i0 != i2) and (i0 != i3) and (i1 != i2) and (i1 != i3) and (i2 != i3):
                    if Gate_type == 'cx':
                        if Gate_pos[0] > Gate_pos[1]:
                            psi_real[i2], psi_real[i3] = psi_real[i3].clone(), psi_real[i2].clone()
                            psi_imag[i2], psi_imag[i3] = psi_imag[i3].clone(), psi_imag[i2].clone()
                        else:
                            psi_real[i1], psi_real[i3] = psi_real[i3].clone(), psi_real[i1].clone()
                            psi_imag[i1], psi_imag[i3] = psi_imag[i3].clone(), psi_imag[i1].clone()
                    elif Gate_type == 'swap':
                        psi_real[i1], psi_real[i2] = psi_real[i2].clone(), psi_real[i1].clone()
                        psi_imag[i1], psi_imag[i2] = psi_imag[i2].clone(), psi_imag[i1].clone()
                    elif Gate_type == 'iswap':
                        psi_real[i1], psi_imag[i1], psi_real[i2], psi_imag[i2] = -psi_imag[i2].clone(), psi_real[i2].clone(), -psi_imag[i1].clone(), psi_real[i1].clone()
                    elif Gate_type == 'cz':
                        psi_real[i3], psi_imag[i3] = -psi_real[i3].clone(), -psi_imag[i3].clone()
                    elif Gate_type == 'cp':
                        angle = torch.tensor(Angles[0])
                        cos_val, sin_val = torch.cos(angle), torch.sin(angle)
                        psi_real[i3], psi_imag[i3] = complex_mul(psi_real[i3], psi_imag[i3], cos_val, sin_val)
                    elif Gate_type == 'syc':
                        psi_real[i1], psi_imag[i1], psi_real[i2], psi_imag[i2] = -psi_imag[i2].clone(), psi_real[i2].clone(), -psi_imag[i1].clone(), psi_real[i1].clone()
                        exp_val = complex_exp(torch.tensor(0.), torch.tensor(-math.pi / 6))
                        psi_real[i3], psi_imag[i3] = complex_mul(psi_real[i3], psi_imag[i3], exp_val[0], exp_val[1])
                    elif Gate_type in ['rxx', 'ryy', 'rzz']:
                        angle = torch.tensor(Angles[0])
                        cos_val, sin_val = torch.cos(angle / 2), torch.sin(angle / 2)
                        # print(cos_val, sin_val)
                        if Gate_type == 'rxx':
                            # print("\nbefore: ", psi_real, psi_imag)
                            psi_real[i0], psi_imag[i0], psi_real[i3], psi_imag[i3] = complex_mul(cos_val, torch.tensor(0.), psi_real[i0], psi_imag[i0])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i3], psi_imag[i3])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i0], psi_imag[i0])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i3], psi_imag[i3])[1], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i3], psi_imag[i3])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i0], psi_imag[i0])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i3], psi_imag[i3])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i0], psi_imag[i0])[1]

                            psi_real[i1], psi_imag[i1], psi_real[i2], psi_imag[i2] = complex_mul(cos_val, torch.tensor(0.), psi_real[i1], psi_imag[i1])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i2], psi_imag[i2])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i1], psi_imag[i1])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i2], psi_imag[i2])[1], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i2], psi_imag[i2])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i1], psi_imag[i1])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i2], psi_imag[i2])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i1], psi_imag[i1])[1]
                            # print("after: ", psi_real, psi_imag)
                            # print(psi_real[i1], psi_imag[i1], psi_real[i2], psi_imag[i2])
                        elif Gate_type == 'ryy':
                            psi_real[i0], psi_imag[i0], psi_real[i3], psi_imag[i3] = complex_mul(cos_val, torch.tensor(0.), psi_real[i0], psi_imag[i0])[0] + complex_mul(torch.tensor(0.), sin_val, psi_real[i3], psi_imag[i3])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i0], psi_imag[i0])[1] + complex_mul(torch.tensor(0.), sin_val, psi_real[i3], psi_imag[i3])[1], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i3], psi_imag[i3])[0] + complex_mul(torch.tensor(0.), sin_val, psi_real[i0], psi_imag[i0])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i3], psi_imag[i3])[1] + complex_mul(torch.tensor(0.), sin_val, psi_real[i0], psi_imag[i0])[1]
                            psi_real[i1], psi_imag[i1], psi_real[i2], psi_imag[i2] = complex_mul(cos_val, torch.tensor(0.), psi_real[i1], psi_imag[i1])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i2], psi_imag[i2])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i1], psi_imag[i1])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i2], psi_imag[i2])[1], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i2], psi_imag[i2])[0] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i1], psi_imag[i1])[0], \
                                                                                    complex_mul(cos_val, torch.tensor(0.), psi_real[i2], psi_imag[i2])[1] + complex_mul(torch.tensor(0.), -sin_val, psi_real[i1], psi_imag[i1])[1]
                        elif Gate_type == 'rzz':
                            exp_pos = complex_exp(torch.tensor(0.), -angle / 2)
                            exp_neg = complex_exp(torch.tensor(0.), angle / 2)
                            psi_real[i0], psi_imag[i0] = complex_mul(psi_real[i0], psi_imag[i0], exp_pos[0], exp_pos[1])
                            psi_real[i1], psi_imag[i1] = complex_mul(psi_real[i1], psi_imag[i1], exp_neg[0], exp_neg[1])
                            psi_real[i2], psi_imag[i2] = complex_mul(psi_real[i2], psi_imag[i2], exp_neg[0], exp_neg[1])
                            psi_real[i3], psi_imag[i3] = complex_mul(psi_real[i3], psi_imag[i3], exp_pos[0], exp_pos[1])

    return psi_real + 1j*psi_imag



if __name__ == '__main__':
    psi_real = torch.tensor([0, 0, 0, 1, 0, 0, 0, 0], dtype=torch.float32)
    psi_imag = torch.tensor([0, 0, 0, 0, 0, 0, 0, 0], dtype=torch.float32)
    num_qubits = 3
    circuit = QuantumCircuit(num_qubits)
    circuit.add_single_gate(0, 'x')
    circuit.add_single_gate(1, 'x')
    circuit.add_double_gate(0, 1, 'cx')
    circuit.add_double_gate(0, 1, 'swap')
    circuit.add_double_gate(0, 1, 'iswap')
    circuit.add_double_gate(1, 2, 'cz')
    circuit.add_double_gate(1, 2, 'cp', 1.78)
    circuit.add_double_gate(0, 1, 'syc')
    circuit.add_double_gate(1, 2, 'rxx', 1.78)
    circuit.add_double_gate(1, 2, 'ryy', 1.78)
    circuit.add_double_gate(1, 2, 'rzz', 1.78)
    # print(circuit.gate_list)
    circuit.run_statevector()
    print(circuit.state)
    # gate_list = [([0, 1], ('cx',)), ([0, 1], ('swap',)), ([0, 1], ('iswap',)), ([1, 2], ('cz', )), ([1, 2], ('cp', 1.78)), ([0, 1], ('syc', )), 
    #              ([1, 2], ('rxx', 1.78)), ([1, 2], ('ryy', 1.78)), ([1, 2], ('rzz', 1.78))]
    gate_list = [([0, 1], ('cx',)), ([0, 1], ('swap',)), ([0, 1], ('iswap',)), ([1, 2], ('cz', )), ([1, 2], ('cp', 1.78)), ([0, 1], ('syc', )),
                 ([1, 2], ('rxx', 1.78)), ([1, 2], ('ryy', 1.78)), ([1, 2], ('rzz', 1.78))]
    for ele in gate_list:
        Gate_type = ele[1][0]
        Gate_pos = ele[0]
        Angles = tuple()
        if len(ele[1]) > 1:
            Angles = tuple(ele[1][1:])
        psi_real, psi_imag = ActOn_State(psi_real, psi_imag, num_qubits, Gate_type, Gate_pos, *Angles)
    final_psi = psi_real + 1j * psi_imag
    print("final_psi:", final_psi)




"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/9/11 14:41
@Function: common_merges.py
@Contact: cuijinghao@tgqs.net
"""
import numpy as np
import re
from tgqSim.circuit.common_gates import BASE_SINGLE_GATE, BASE_DOUBLE_GATE, BASE_TRIPLE_GATE
from tgqSim.circuit.common_gates import _CommonGate as CommonGate
from tgqSim.circuit.common_gates import (x, rz, cz, sqrt_x)
from tgqSim.circuit.QuantumCircuit import QuantumCircuit
from tgqSim.transformers.common_decompositions import *

# a set of gates among which, if encountered twice in a circuit on a same qbit, can be merged
GATE_MERGE_SET = ['sqrt_x', 'rz', 'cx']

# decomposition base gates set
DECOMP_GATE_SET = {'rz', 'cz', 'x', 'sqrt_x'}


def two_gates_can_merge(gate1: CommonGate, gate2: CommonGate) -> bool:
    """

    Args:
        gate1:
        gate2:

    Returns: compare two gates' position and check gate type

    """
    return (gate1.name in GATE_MERGE_SET
            and gate2.name in GATE_MERGE_SET
            and gate1.name == gate2.name
            and gate1.qbit == gate2.qbit)


def merge_two_gates_on_base(gate1: CommonGate, gate2: CommonGate):
    """
    Params: original two single bit gates
    Returns: gate after merging

    """
    if two_gates_can_merge(gate1, gate2):
        if gate1.name.lower() == 'sqrt_x' and gate2.name.lower() == 'sqrt_x':
            return x(gate1.qbit)
        elif gate1.name.lower() == 'x' and gate2.name.lower() == 'x':
            return None
        elif gate1.name.lower() == 'cx' and gate2.name.lower() == 'cx':
            return None
        elif gate1.name.lower() == 'rz' and gate2.name.lower() == 'rz':
            return rz(gate1.qbit, gate1.theta + gate2.theta)


def merge_circuit():
    """
    Params: original circuit, merge function
    Returns: merged circuit
    Description: compare each two gates, if can_merge then merge these two gates

    """
    pass


def decompose_circuit_on_base(circuit: QuantumCircuit) -> QuantumCircuit:
    """

    Args:
        circuit: original circuit

    Returns:
        circuit after decomposition

    """
    org_width = circuit.width
    decomposed_circuit = QuantumCircuit(org_width)
    for gate in circuit.gate_list:
        # [(1, ('rx', 1.5707963267948966)), (0, ('sqrt_x',))]
        gate_name = gate[1][0].lower()
        if gate_name in BASE_SINGLE_GATE:
            gate_pos = gate[0]
            #### base single gates
            if gate_name == 'x':
                decomposed_circuit.append(x(gate_pos))
            elif gate_name == 'rz':
                decomposed_circuit.append(rz(gate_pos, gate[1][1]))
            #####
            elif gate_name == 'h':
                decomposed_circuit.append_circuit(H(gate_pos, org_width))
            elif gate_name == 'y':
                decomposed_circuit.append_circuit(Y(gate_pos, org_width))
            elif gate_name == 'z':
                decomposed_circuit.append_circuit(Z(gate_pos, org_width))
            elif gate_name == 's':
                decomposed_circuit.append_circuit(S(gate_pos, org_width))
            elif gate_name == 'sdg':
                decomposed_circuit.append_circuit(Sdg(gate_pos, org_width))
            elif gate_name == 't':
                decomposed_circuit.append_circuit(T(gate_pos, org_width))
            elif gate_name == 'tdg':
                decomposed_circuit.append_circuit(Tdg(gate_pos, org_width))
            elif gate_name == 'rx':
                decomposed_circuit.append_circuit(Rx(gate[1][1], gate_pos, org_width))
            elif gate_name == 'ry':
                decomposed_circuit.append_circuit(Ry(gate[1][1], gate_pos, org_width))

        elif gate_name in BASE_DOUBLE_GATE:
            pos1, pos2 = gate[0][0], gate[0][1]
            #### base double gates
            if gate_name == 'cz':
                decomposed_circuit.append(cz(pos1, pos2))
            ####
            elif gate_name == 'cx':
                decomposed_circuit.append_circuit(Cnot(pos1, pos2, org_width))
            elif gate_name == 'swap':
                decomposed_circuit.append_circuit(Swap(pos1, pos2, org_width))
            elif gate_name == 'iswap':
                decomposed_circuit.append_circuit(ISwap(pos1, pos2, org_width))
            elif gate_name == 'cp':
                decomposed_circuit.append_circuit(Cphase(gate[1][1], pos1, pos2, org_width))
            elif gate_name == 'syc':
                decomposed_circuit.append_circuit(Syc(pos1, pos2, org_width))
            elif gate_name == 'rxx':
                decomposed_circuit.append_circuit(Rxx(gate[1][1], pos1, pos2, org_width))
            elif gate_name == 'ryy':
                decomposed_circuit.append_circuit(Ryy(gate[1][1], pos1, pos2, org_width))
            elif gate_name == 'rzz':
                decomposed_circuit.append_circuit(Rzz(gate[1][1], pos1, pos2, org_width))
        elif gate_name in BASE_TRIPLE_GATE:
            # cswap := fredkin
            if gate_name == 'cswap':
                decomposed_circuit.append_circuit(Fred(gate[0][0], gate[0][1], gate[0][2], org_width))
            # ccx := toffoli
            elif gate_name == 'ccx':
                decomposed_circuit.append_circuit(Toff(gate[0][0], gate[0][1], gate[0][2], org_width))
    return decomposed_circuit



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/6/28 15:49
@Function: common_gates
@Contact: cuijinghao@tgqs.net
"""
from typing import Union
from abc import abstractmethod
from copy import deepcopy

# global parameters
BASE_SINGLE_GATE = ['h', 'x', 'y', 'z', 'rx', 'ry', 'rz', 'u3', 's', 'sdg', 't', 'tdg', "damp_I", "pd", "ad", 'sqrt_x']
BASE_DOUBLE_GATE = ['cx', 'swap', 'iswap', 'cz', 'cp', 'rxx', 'ryy', 'rzz', 'syc']
BASE_TRIPLE_GATE = ['ccx', 'cswap']
CONTROLLED_GATE = ['cx', 'cz', 'cp', 'ccx', 'cswap']
ROTATION_GATE = ['rxx', 'ryy', 'rzz']
MEASURE = ['measure']
BASE_SINGLE_GATE_MAP = {'h': '_H', 'x': '_X', 'y': '_Y', 'z': '_Z',
                        'rx': '_RX', 'ry': '_RY', 'rz': '_RZ',
                        'u3': '_U3',
                        's': '_S', 'sdg': '_SDG', 't': '_T', 'tdg': '_TDG', 'sqrt_x': '_SQRT_X'}
BASE_DOUBLE_GATE_MAP = {'cx': '_CX', 'swap': '_SWAP', 'iswap': '_ISWAP',
                        'cz': '_CZ', 'cp': '_CP', 'rxx': '_RXX', 'ryy': '_RYY', 'rzz': '_RZZ', 'syc': '_SYC'}

BASE_TRIPLE_GATE_MAP = {'cswap': '_CSWAP', 'ccx': '_CCX'}

MEASURE_MAP = {'measure': '_Measure'}


# will implement more gates classes later
# will further abstract classes
class _CommonGate:
    def __init__(self, qbit=None, name=None, theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
    @abstractmethod
    def gate(self, *args, **kwargs):
        gate_instance = deepcopy(self)
        return gate_instance


CommonGate = _CommonGate()


class _Measure(_CommonGate):
    def __init__(self, qbit: Union[list, int] = 0, name: str = 'measure', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        if isinstance(qbit, int):
            self.display_name = ('Measure-z',)
        else:
            self.display_name = tuple(['Measure-z'] * len(self.qbit))

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        if isinstance(qbit, int):
            gate_instance.display_name = ('Measure-z',)
        else:
            gate_instance.display_name = tuple(['Measure-z'] * len(gate_instance.qbit))
        return gate_instance


measure = _Measure()


class _I(_CommonGate):
    """
    eye matrix
    """
    def __init__(self, qbit: int = 0, name='I', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("I", )

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


I = _I()


class _Pauli(_CommonGate):
    def __init__(self):
        pass

    def __call__(self):
        return

    def gate(self):
        pass


class _X(_Pauli):
    # todo: add attribute noise
    def __init__(self, qbit: int = 0, name='x', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("X", )

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


x = _X()


class _SQRT_X(_Pauli):
    def __init__(self, qbit: int = 0, name='sqrt_x', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("X^½", )

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


sqrt_x = _SQRT_X


class _Y(_Pauli):
    def __init__(self, qbit=0, name='y', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("Y",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


y = _Y()


class _Z(_Pauli):
    def __init__(self, qbit=0, name='z', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("Z",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


z = _Z()


class _RX(_CommonGate):
    def __init__(self, qbit=0, name='rx', theta=0):
        self.qbit = qbit
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Rx({round(self.theta, 2)})",)

    def __call__(self, qbit, theta):
        return self.gate(qbit, theta)

    def gate(self, qbit, theta):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        gate_instance.theta = float(theta)
        gate_instance.display_name = (f"Rx({round(gate_instance.theta, 2)})",)
        return gate_instance


rx = _RX()


class _RY(_CommonGate):
    def __init__(self, qbit=0, name='ry', theta=0):
        self.qbit = qbit
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Ry({round(self.theta, 2)})",)

    def __call__(self, qbit, theta):
        return self.gate(qbit, theta)

    def gate(self, qbit, theta):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        gate_instance.theta = float(theta)
        gate_instance.display_name = (f"Ry({round(gate_instance.theta, 2)})",)
        return gate_instance


ry = _RY()


class _RZ(_CommonGate):
    def __init__(self, qbit=0, name='rz', theta=0):
        self.qbit = qbit
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Rz({round(self.theta, 2)})",)

    def __call__(self, qbit, theta):
        return self.gate(qbit, theta)

    def gate(self, qbit, theta):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        gate_instance.theta = float(theta)
        gate_instance.display_name = (f"Rz({round(gate_instance.theta, 2)})",)
        return gate_instance


rz = _RZ()


# class _U3(_CommonGate):
#     def __init__(self, qbit=0, name='u3', *theta):
#         self.qbit = qbit
#         self.name = name
#         self.theta = theta
#
#     def __call__(self, qbit, *theta):
#
#         # print('len of theta tuple is {}, tuple is {}'.format(len(theta), theta))
#         if not theta or (theta and len(theta) != 3):
#             raise Exception('U3 gate requires 3 parameters, please check!')
#         # print('theta before calling gate is {}'.format(theta))
#         return self.gate(qbit, theta[0], theta[1], theta[2])
#
#     def gate(self, qbit, *theta):
#         # print('theta after initialization is {}'.format(theta))
#         self.qbit = qbit
#         self.theta = theta
#         return self
#
#
# u3 = _U3()


# class

class _H(_CommonGate):
    def __init__(self, qbit=0, name='h', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ('H',)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


h = _H()


class _S(_CommonGate):
    def __init__(self, qbit=0, name='s', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("S",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


s = _S()


class _SDG(_CommonGate):
    def __init__(self, qbit=0, name='sdg', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("SDG",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


sdg = _SDG()


class _T(_CommonGate):
    def __init__(self, qbit=0, name='t', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("T",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


t = _T()


class _TDG(_CommonGate):
    def __init__(self, qbit=0, name='tdg', theta=None):
        self.qbit = qbit
        self.name = name
        self.theta = theta
        self.display_name = ("TDG",)

    def __call__(self, qbit):
        return self.gate(qbit)

    def gate(self, qbit):
        gate_instance = deepcopy(self)
        gate_instance.qbit = qbit
        return gate_instance


tdg = _TDG()


# double gates ['cx', 'swap', 'iswap', 'cz', 'cp', 'rxx', 'ryy', 'rzz', 'syc']
class _CX(_CommonGate):
    """
    @explain: CNOT gate
    @params: by default, qbit0 is the controll bit
    @return: CNOT gate
    """
    # by default, qbit0 is the controll bit
    def __init__(self, qbit0=0, qbit1=1, name='cx', theta=None):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("@", "X")

    def __call__(self, qbit0, qbit1):
        return self.gate(qbit0, qbit1)

    def gate(self, qbit0, qbit1):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        return gate_instance


cx = _CX()


class _CZ(_CommonGate):
    def __init__(self, qbit0=0, qbit1=1, name='cz', theta=None):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("@", "@")

    def __call__(self, qbit0, qbit1, theta=None):
        return self.gate(qbit0, qbit1, theta)

    def gate(self, qbit0, qbit1, theta=None):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        return gate_instance


cz = _CZ()


class _CP(_CommonGate):
    """
    @params: first control, second target, third theta
    @returns: cphase gate
    @explanation: cphase gate

    """
    def __init__(self, qbit0=0, qbit1=1, name='cp', theta=0):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = float(theta)
        self.display_name = ("@", f"@^{round(self.theta, 2)}")

    def __call__(self, qbit0, qbit1, theta):
        return self.gate(qbit0, qbit1, theta)

    def gate(self, qbit0, qbit1, theta):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        gate_instance.theta = float(theta)
        gate_instance.display_name = ("@", f"@^{round(gate_instance.theta, 2)}")
        return gate_instance


cp = _CP()


class _SWAP(_CommonGate):
    def __init__(self, qbit0=0, qbit1=1, name='swap', theta=None):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("x", "x")

    def __call__(self, qbit0, qbit1):
        return self.gate(qbit0, qbit1)

    def gate(self, qbit0, qbit1):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        return gate_instance


swap = _SWAP()


class _ISWAP(_CommonGate):
    def __init__(self, qbit0: int = 0, qbit1: int = 1, name='iswap', theta=None):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("iSwap", "iSwap")

    def __call__(self, qbit0, qbit1):
        return self.gate(qbit0, qbit1)

    def gate(self, qbit0, qbit1):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        return gate_instance


iswap = _ISWAP()


class _RXX(_CommonGate):
    def __init__(self, qbit0=0, qbit1=1, name='rxx', theta=0):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Rxx({round(self.theta, 2)})", f"Rxx({round(self.theta, 2)})")

    def __call__(self, qbit0, qbit1, theta):
        return self.gate(qbit0, qbit1, theta)

    def gate(self, qbit0, qbit1, theta):
        gate_instance = deepcopy(self)
        gate_instance.theta = float(theta)
        gate_instance.qbit = [qbit0, qbit1]
        gate_instance.display_name = (f"Rxx({round(gate_instance.theta, 2)})", f"Rxx({round(gate_instance.theta, 2)})")
        return gate_instance


rxx = _RXX()


class _RYY(_CommonGate):
    def __init__(self, qbit0=0, qbit1=1, name='ryy', theta=0):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Ryy({round(theta, 2)})", f"Ryy({round(theta, 2)})")

    def __call__(self, qbit0, qbit1, theta):
        return self.gate(qbit0, qbit1, theta)

    def gate(self, qbit0, qbit1, theta):
        gate_instance = deepcopy(self)
        gate_instance.theta = float(theta)
        gate_instance.qbit = [qbit0, qbit1]
        gate_instance.display_name = (f"Ryy({round(gate_instance.theta, 2)})", f"Ryy({round(gate_instance.theta, 2)})")
        return gate_instance


ryy = _RYY()


class _RZZ(_CommonGate):
    """
    @params: first two params are qbit positions, third param is theta;
    @return: rzz gate
    """
    def __init__(self, qbit0=0, qbit1=1, name='rzz', theta=0):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = float(theta)
        self.display_name = (f"Rzz({round(self.theta, 2)})", f"Rzz({round(self.theta, 2)})")

    def __call__(self, qbit0, qbit1, theta):
        return self.gate(qbit0, qbit1, theta)

    def gate(self, qbit0, qbit1, theta):
        gate_instance = deepcopy(self)
        gate_instance.theta = float(theta)
        gate_instance.qbit = [qbit0, qbit1]
        gate_instance.display_name = (f"Rzz({round(gate_instance.theta, 2)})", f"Rzz({round(gate_instance.theta, 2)})")
        return gate_instance


rzz = _RZZ()


class _SYC(_CommonGate):
    def __init__(self, qbit0: int = 0, qbit1: int = 1, name='syc', theta=None):
        self.qbit0 = qbit0
        self.qbit1 = qbit1
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ('SYC', 'SYC')

    def __call__(self, qbit0, qbit1):
        return self.gate(qbit0, qbit1)

    def gate(self, qbit0, qbit1):
        gate_instance = deepcopy(self)
        gate_instance.qbit = [qbit0, qbit1]
        return gate_instance


syc = _SYC()


class _CCX(_CommonGate):
    """
    @params: control_qbit control qbits: list, target_qbit target qbit: int
    """
    def __init__(self, control_qbit: Union[list, int] = None, target_qbit: Union[list, int] = None, name='ccx', theta=None):
        self.control_qbit = control_qbit if isinstance(control_qbit, list) else [control_qbit]
        self.target_qbit = target_qbit if isinstance(target_qbit, list) else [target_qbit]
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("@", "@", "X")

    def __call__(self, control_qbit, target_qbit):
        return self.gate(control_qbit, target_qbit)

    def gate(self, control_qbit, target_qbit):
        gate_instance = deepcopy(self)
        gate_instance.control = control_qbit if isinstance(control_qbit, list) else [control_qbit]
        gate_instance.target = target_qbit if isinstance(target_qbit, list) else [target_qbit]
        gate_instance.qbit = gate_instance.control + gate_instance.target
        return gate_instance


ccx = _CCX()


class _CSWAP(_CommonGate):
    """
    @params: control_qbit control qbit: int, target_qbit target qbits: list
    """
    def __init__(self, control_qbit: Union[list, int] = None, target_qbit: Union[list, int] = None, name='cswap', theta=None):
        self.control_qbit = control_qbit if isinstance(control_qbit, list) else [control_qbit]
        self.target_qbit = target_qbit if isinstance(target_qbit, list) else [target_qbit]
        self.qbit = None
        self.name = name
        self.theta = theta
        self.display_name = ("@", "x", "x")

    def __call__(self, control_qbit, target_qbit):
        return self.gate(control_qbit, target_qbit)

    def gate(self, control_qbit, target_qbit):
        gate_instance = deepcopy(self)
        gate_instance.control = control_qbit if isinstance(control_qbit, list) else [control_qbit]
        gate_instance.target = target_qbit if isinstance(target_qbit, list) else [target_qbit]
        gate_instance.qbit = gate_instance.control + gate_instance.target
        return gate_instance


cswap = _CSWAP()






"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/7/17 16:24
@Function: logger.py
@Contact: cuijinghao@tgqs.net
"""
import logging


class Logger:
    def __init__(self, name: str, level: int = logging.INFO, log_file: str = None):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)

        # Create a console handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)

        # Create a formatter and set it for the console handler
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(formatter)

        # Add the console handler to the logger
        self.logger.addHandler(console_handler)

        # If a log file is specified, create a file handler
        if log_file:
            file_handler = logging.FileHandler(log_file)
            file_handler.setLevel(level)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

    def debug(self, message: str):
        self.logger.debug(message)

    def info(self, message: str):
        self.logger.info(message)

    def warning(self, message: str):
        self.logger.warning(message)

    def error(self, message: str):
        self.logger.error(message)

    def critical(self, message: str):
        self.logger.critical(message)


if __name__ == "__main__":
    log = Logger(name="tgqsim", level=logging.DEBUG, log_file="app.log")
    log.debug("This is a debug message")
    log.info("This is an info message")
    log.warning("This is a warning message")
    log.error("This is an error message")
    log.critical("This is a critical message")



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/7/18 15:25
@Function: qasmParser.py
@Contact: cuijinghao@tgqs.net
"""


import re
from tgqSim.circuit.QuantumCircuit import QuantumCircuit as qc
from tgqSim.circuit.common_gates import (BASE_SINGLE_GATE,
                                               BASE_DOUBLE_GATE,
                                               BASE_TRIPLE_GATE,
                                               MEASURE,
                                               BASE_SINGLE_GATE_MAP,
                                               BASE_DOUBLE_GATE_MAP,
                                               BASE_TRIPLE_GATE_MAP)
import importlib
import os
import sys

sys.path.append(os.path.join(os.path.dirname(os.getcwd())))
# MODULE = importlib.import_module('..common_gates', 'circuit.common_gates')
MODULE = importlib.import_module('.common_gates', 'tgqSim.circuit')
# print(MODULE)
QASM_REG = ['qreg']
QASM_FORMAT = ['openqasm']


class QASMParser:
    def __init__(self, qasm_code: str):
        self.qasm_code = qasm_code
        self.lines = self.qasm_code.split('\n')
        self.circuit = qc()
        # self.qubits = {}
        # self.bits = {}

    def parse(self):
        # todo: receives an ast string and returns a quantumcircuit object
        # todo: add more operations like if and or in the future
        measure_list = []
        for line in self.lines:
            line = line.strip()
            # todo: change regex method
            # todo: regex for u3 gate
            # get operation from the line
            # this regex get the command
            # e.g. qreg q[5] - > qreg; u3(0.5,0.5,0.5) q[0] -> u3(0.5,0.5,0.5);
            # command = re.match(r'^([a-zA-Z]+\(?[-\d.]*\)?)', line)
            command = re.match(r'^([a-zA-Z]+\d*)(\(?[^)]*\)?)', line)

            qbits = list(map(int, re.findall(r'q\[(\d+)\]', line)))
            # if command:
            #     print('command is : {}'.format(command.group(1)))
            if command and qbits:
                command = command.group(1).lower()
                print('command initial is {}'.format(command))
                # check whether command has theta parameter, e.g. rz(pi)

                theta_flag = re.match(r'^([a-zA-Z]+\d*)\(([^)]+)\)', line)
                theta = None
                # print('theta flag {}'.format(theta_flag))
                if theta_flag:
                    # command = theta_flag.group(1)
                    # e.g. str: 0.708, 0.708, 0.708
                    theta = theta_flag.group(2)
                    lst = []
                    for _ in theta.split(','):
                        lst.append(float(_.strip()))
                    theta = tuple(lst)

                # print('theta after theta checking is {}'.format(theta))
                # print('theta type after theta checking is {}'.format(type(theta)))
                # print('command after theta checking is {}'.format(command))
                if command in QASM_FORMAT:
                    continue
                elif command in QASM_REG:
                    # e.g. qreg q[10]
                    self.circuit.add_qubits(qbits[0])
                elif command in BASE_SINGLE_GATE_MAP.keys():
                    gate = getattr(MODULE, BASE_SINGLE_GATE_MAP[command])
                    # instantiate
                    gate = gate()
                    if theta:
                        # u1
                        if len(theta) == 1:
                            self.circuit.append(gate(qbits[0], theta[0]))
                        # u2
                        elif len(theta) == 2:
                            self.circuit.append(gate(qbits[0], theta[0], theta[1]))
                        # u3
                        elif len(theta) == 3:
                            self.circuit.append(gate(qbits[0], theta[0], theta[1], theta[2]))
                    else:
                        self.circuit.append(gate(qbits[0]))
                elif command in BASE_DOUBLE_GATE_MAP.keys():
                    gate = getattr(MODULE, BASE_DOUBLE_GATE_MAP[command])
                    gate = gate()
                    if theta:
                        self.circuit.append(gate(qbits[0], qbits[1], theta))
                    else:
                        self.circuit.append(gate(qbits[0], qbits[1]))
                elif command in BASE_TRIPLE_GATE_MAP.keys():
                    pass
                elif command in MEASURE:
                    measure_list += qbits
        if measure_list and measure_list != []:
            self.circuit.add_measure_gate(measure_list, 'measure')

        return self.circuit

    def _parse_measurement(self, line):
        match = re.match(r'measure (\w+) -> (\w+);', line)
        if match:
            qubit_name = match.group(1)
            bit_name = match.group(2)
            self.circuit.append(f"Measure {qubit_name} to {bit_name}")

    def _parse_if_statement(self, line):
        # 示例解析if语句
        pass

    def _parse_for_loop(self, line):
        # 示例解析for循环
        pass

    def _parse_function_definition(self, line):
        # 示例解析函数定义
        pass

# qasm_code =  """
# OPENQASM 2.0;
# qreg q[3];
# h q[0];
# cx q[0], q[1];
# toffoli q[0], q[1], q[2];
# rz(0.708) q[0];
# rz(0) q[0];
# rz(-0.5) q[0];
# u3(0.708, 0.708, 0.708) q[1];
# measure q[0] ;
# measure q[1] ;
# """
# print(QASMParser(qasm_code).parse().gate_list)




"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/8/22 15:25
@Function: qasmGenerator.py
@Contact: cuijinghao@tgqs.net
"""
from pygments.lexer import RegexLexer
from pygments.token import Comment, String, Keyword, Name, Number, Text
from pygments.style import Style
from pygments.formatters import Terminal256Formatter
import pygments


class QasmTerminalStyle(Style):
    """A style for OpenQasm in a Terminal env (e.g. Jupyter print)."""

    styles = {
        String: "ansibrightred",
        Number: "ansibrightcyan",
        Keyword.Reserved: "ansibrightgreen",
        Keyword.Declaration: "ansibrightgreen",
        Keyword.Type: "ansibrightmagenta",
        Name.Builtin: "ansibrightblue",
        Name.Function: "ansibrightyellow",
    }


class QasmHTMLStyle(Style):
    """A style for OpenQasm in a HTML env (e.g. Jupyter widget)."""

    styles = {
        String: "ansired",
        Number: "ansicyan",
        Keyword.Reserved: "ansigreen",
        Keyword.Declaration: "ansigreen",
        Keyword.Type: "ansimagenta",
        Name.Builtin: "ansiblue",
        Name.Function: "ansiyellow",
    }


class OpenQASMLexer(RegexLexer):
    """A pygments lexer for OpenQasm."""

    name = "OpenQASM"
    aliases = ["qasm"]
    filenames = ["*.qasm"]

    gates = [
        "id",
        "cx",
        "x",
        "y",
        "z",
        "s",
        "sdg",
        "h",
        "t",
        "tdg",
        "ccx",
        "c3x",
        "c4x",
        "c3sqrtx",
        "rx",
        "ry",
        "rz",
        "cz",
        "cy",
        "ch",
        "cp",
        "swap",
        "syc",
        "cswap",
        "crx",
        "cry",
        "crz",
        "cu1",
        "cu3",
        "rxx",
        "rzz",
        "rccx",
        "rc3x",
        "u1",
        "u2",
        "u3",
    ]

    tokens = {
        "root": [
            (r"\n", Text),
            (r"[^\S\n]+", Text),
            (r"//\n", Comment),
            (r"//.*?$", Comment.Single),
            # Keywords
            (r"(OPENQASM|include)\b", Keyword.Reserved, "keywords"),
            (r"(qreg|creg)\b", Keyword.Declaration),
            # Treat 'if' special
            (r"(if)\b", Keyword.Reserved, "if_keywords"),
            # Constants
            (r"(pi)\b", Name.Constant),
            # Special
            (r"(barrier|measure|reset)\b", Name.Builtin, "params"),
            # Gates (Types)
            ("(" + "|".join(gates) + r")\b", Keyword.Type, "params"),
            (r"[unitary\d+]", Keyword.Type),
            # Functions
            (r"(gate)\b", Name.Function, "gate"),
            # Generic text
            (r"[a-zA-Z_][a-zA-Z0-9_]*", Text, "index"),
        ],
        "keywords": [
            (r'\s*("([^"]|"")*")', String, "#push"),
            (r"\d+", Number, "#push"),
            (r".*\(", Text, "params"),
        ],
        "if_keywords": [
            (r"[a-zA-Z0-9_]*", String, "#pop"),
            (r"\d+", Number, "#push"),
            (r".*\(", Text, "params"),
        ],
        "params": [
            (r"[a-zA-Z_][a-zA-Z0-9_]*", Text, "#push"),
            (r"\d+", Number, "#push"),
            (r"(\d+\.\d*|\d*\.\d+)([eEf][+-]?[0-9]+)?", Number, "#push"),
            (r"\)", Text),
        ],
        "gate": [(r"[unitary\d+]", Keyword.Type, "#push"), (r"p\d+", Text, "#push")],
        "index": [(r"\d+", Number, "#pop")],
    }


def qasm(gate_list: list, nqubit: int, formatted=False)->str:
    mapping = {
        "iswap": "gate iswap q0,q1 { s q0; s q1; h q0; cx q0,q1; cx q1,q0; h q1; }",
        "ryy": "gate ryy(param0) q0,q1 { rx(pi/2) q0; rx(pi/2) q1; cx q0,q1; rz(param0) q1; cx q0,q1; rx(-pi/2) q0; rx(-pi/2) q1; }",
        "syc": "gate syc q0,q1 { rxx(-pi/2) q0,q1; ryy(-pi/2) q0,q1; cp(-pi/6) q0,q1; }"
    }
    version_lib = "OPENQASM 2.0;\ninclude \"qelib1.inc\";\n"
    bytes_pos = [f"qreg q[{nqubit}]", f"creg c[{nqubit}]"]
    other_gate_define = []
    normal_gate_list = []
    is_define = {"iswap": False, "ryy": False, "syc": False}
    measure_gate = []
    for pos, gate_info in gate_list:
        gate_name = gate_info[0].lower()
        if gate_name == "measure":
            if isinstance(pos, int):
                measure_gate.append(f"{gate_name} q[{pos}] -> c[{pos}]")
            elif isinstance(pos, list):
                for pos_idx in pos:
                    measure_gate.append(f"{gate_name} q[{pos_idx}] -> c[{pos_idx}]")
            continue
        if isinstance(pos, int):
            normal_gate_list.append(f"{gate_name} q[{pos}]")
        elif isinstance(pos, list):
            length = len(pos)
            if 1 == length:
                normal_gate_list.append(f"{gate_info[0].lower()} q[{pos[0]}]")
            elif 2 == length:
                need_redefine_gate = ["iswap", "ryy", "syc"]
                if gate_name in need_redefine_gate and not is_define[gate_name]:
                    other_gate_define.append(mapping[gate_name])
                    is_define[gate_name] = True
                if len(gate_info) >= 2:
                    if isinstance(gate_info[1], float):
                        normal_gate_list.append(f"{gate_name}({gate_info[1]}) q[{pos[0]}] q[{pos[1]}]")
                    else:
                        normal_gate_list.append(f"{gate_name} q[{pos[0]}] q[{pos[1]}]")
            elif 3 == length:
                if len(gate_info) >= 2:
                    if isinstance(gate_info[1], float):
                        normal_gate_list.append(f"{gate_name}({gate_info[1]}) q[{pos[0]}] q[{pos[1]}]")
                    else:
                        normal_gate_list.append(f"{gate_name} q[{pos[0]}] q[{pos[1]}]")
    qasm_code = version_lib + "\n".join(other_gate_define) + "\n" + "\n".join(bytes_pos) + "\n" + \
        "\n".join(normal_gate_list) + "\n" + "\n".join(measure_gate)
    if formatted:
        code = pygments.highlight(
            qasm_code,
            OpenQASMLexer(),
            Terminal256Formatter(style=QasmTerminalStyle)
        )
        print(code)
        return None
    return qasm_code



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/7/2 14:37
@Function: test
@Contact: cuijinghao@tgqs.net
"""
from circuit.QuantumCircuit import QuantumCircuit
import numpy as np
from sim.QuantumSimulator import QuantumSimulator
from openqasm2.qasmParser import QASMParser
import tgqSim as tgqs
from tgqSim.utils.visualization import to_text_diag

if __name__ == '__main__':
    # use case as below
    nQubits = 3
    qc = QuantumCircuit(0)
    qc.add_qubits(nQubits, name='qft')
    hgatetest = tgqs.h
    hgatetest(1)
    # # hgate = tgqs.h(1)
    # # # print(hgate)
    # qc.append(hgatetest)
    # print(qc.gate_list)

    u3gatetest = tgqs.u3
    u3gatetest(1, 0.1, 0.2, 0.3)
    # qc.append(u3gatetest)
    # print(qc.gate_list)



    ###########
    x0 = tgqs.x(0)
    x1 = tgqs.x(1)
    z0 = tgqs.z(0)
    z1 = tgqs.z(1)
    print(id(z0), id(z1))
    qc.append(z0)
    qc.append(z1)
    print(qc.gate_list)






    xgate = tgqs.x(0)
    # print(xgate)
    # qc.append(xgate)
    # print(qc)
    # qc.with_noise(0,'depolarize', 0.41)

    #

    # qc.append(tgqs.measure([0]))
    # print(qc.noise_circuit)
    # print(qc.displayname_list)
    # print(qc)

    # print(qc.gate_list)
    # rxgate = tgqs.rx(qbit=1, theta=0.707)
    # # print(rxgate)
    # qc.append(rxgate)
    # print(qc.gate_list)
    # rxxgate = tgqs.rxx(0,1,0.707)
    # qc.append(rxxgate)
    # print(qc.gate_list)
    # rxgatetest = tgqs.rx
    # rxgatetest(1, 0.707)
    # qc.append(rxgatetest)
    #
    # ccxgatetest = tgqs.ccx
    # ccxgatetest([0,2], 1)
    #
    # cxgatetest = tgqs.cx
    # cxgatetest()
    # cpgatetest = tgqs.cp
    # cpgatetest()
    #
    # tgqs.rxx()

    # measuregatetest = tgqs.measure

    # tgqs.ccx()
    # tgqs.rzz()


    # qc.append(ccxgatetest)
    # print(qc.gate_list)
    # print(qc)

    # qasm_code = """
    # OPENQASM 2.0;
    # qreg q[5]
    # h q[0];
    # cx q[0], q[1];
    # rz(0.708) q[0];
    # rz(0) q[0];
    # rz(-0.5) q[0];
    # u3(0.708, 0.708, 0.708) q[1];
    #
    # """
    # qc = QASMParser(qasm_code).parse()
    # print(qc)
    # # qc.append(tgqs.measure([2,1,0,3]))
    # print(qc.width)
    # print(qc.gate_list)
    # qc.with_noise(0,'bit_flip', 0.1)
    # print(qc.noise_circuit)
    # print(qc.gate_list)
    # simulator = QuantumSimulator()
    # noise_res = simulator.run_with_noise(qc, shots=1000)
    # print(noise_res)

    # # # get state vector
    # simulated_state = simulator.run_statevector(qc, 'CPU')
    # print(simulated_state)
    # #
    # # # get probability vector
    # result = simulator.execute(qc)
    # print(result)
    # prob = simulator.freq_to_prob(result)
    # print(prob)
    # qc.add_single_gate(hgatetest)
    # qc.append()
    # for i in range(nQubits):
    #     if nQubits - 2 == i:
    #         qc.x(i)
    #     else:
    #         qc.h(i)
    # for i in range(nQubits - 1, -1, -1):
    #     for j in range(i, -1, -1):
    #         if j == i:
    #             qc.h(j)
    #         else:
    #             qc.cp(control_qubit=j, target_qubit=i, theta=np.pi / (2 ** (i - j)))
    # qc.x(0)
    # for i in range(0, nQubits // 2):
    #     qc.swap(qubit_1=i, qubit_2=nQubits - 1 - i)
    #
    # simulator = QuantumSimulator()
    # simulator.
    # qc.run_statevector()
    # print(qc.state)
    # print(len(qc.state))
    # measure_pos = sorted([1, 0], reverse=True)
    # print(qc.execute(measure_bits_list=measure_pos))
    # qc.show_quantum_circuit()



"""
@author: Yuchen He
@contact: heyuchen@tgqs.net
@version: 1.0.0
@file: DoubleGate.py
@time: 2024/1/16 17:05
"""
import numpy as np
from numba import njit, prange
import cmath as cm


def ActOn_State(psi, num_qubits, Gate_type, Gate_pos, *Angles):
    # Gate_pos[0]:control bit
    j1 = sorted(Gate_pos)[1]  # 高位比特，不区分控制位与目标位
    # Gate_pos[1]:target bit
    j0 = sorted(Gate_pos)[0]  # 低位比特，不区分控制位与目标位

    i1_plus = 2 ** j0  #
    i2_plus = 2 ** j1  #
    i3_plus = 2 ** j0 + 2 ** j1  # for i3 = i0 + i3_plus
    delta2 = 2 ** (j1 + 1)
    delta1 = 2 ** (j0 + 1)
    max2 = 2 ** num_qubits - delta2
    max1 = 2 ** j1 - delta1
    max0 = 2 ** j0 - 1

    for k in prange(0, max2 + 1, delta2):
        for l in prange(0, max1 + 1, delta1):
            for m in prange(0, max0 + 1):
                i0 = m | l | k  # to get a(*...*0_j1*...*0_j0*...)
                i1 = i0 | i1_plus  # to get a(*...*0_j1*...*1_j0*...)
                i2 = i0 | i2_plus  # to get a(*...*1_j1*...*0_j0*...)
                i3 = i0 | i3_plus  # to get a(*...*1_j1*...*1_j0*...)
                if Gate_type == 'cx':
                    if Gate_pos[0] > Gate_pos[1]:
                        psi[i2],  psi[i3] =  psi[i3],  psi[i2]
                    else:
                        psi[i1],  psi[i3] =  psi[i3],  psi[i1]
                elif Gate_type == 'swap':
                    psi[i1], psi[i2] = psi[i2],  psi[i1]
                elif Gate_type == 'iswap':
                    psi[i1], psi[i2] = 1j * psi[i2], 1j * psi[i1]
                elif Gate_type == 'cz':
                    psi[i3] = -1.0 * psi[i3]
                elif Gate_type == 'cp':
                    psi[i3] = cm.exp(1j * Angles[0]) * psi[i3]
                elif Gate_type == 'syc':
                    psi[i1], psi[i2], psi[i3] = -1j * psi[i2], -1j * psi[i1], cm.exp(-1j * np.pi / 6) * psi[i3]
                elif Gate_type == 'rxx':
                    psi[i0], psi[i3] = cm.cos(Angles[0] / 2.0) *  psi[i0] + (-1j) * cm.sin(
                        Angles[0] / 2.0) *  psi[i3], \
                                                cm.cos(Angles[0] / 2.0) *  psi[i3] + (-1j) * cm.sin(
                                                    Angles[0] / 2.0) *  psi[i0]
                    psi[i1], psi[i2] = cm.cos(Angles[0] / 2.0) * psi[i1] + (-1j) * cm.sin(
                        Angles[0] / 2.0) *  psi[i2], \
                                                cm.cos(Angles[0] / 2.0) * psi[i2] + (-1j) * cm.sin(
                                                    Angles[0] / 2.0) * psi[i1]
                elif Gate_type == 'ryy':
                    psi[i0], psi[i3] = cm.cos(Angles[0] / 2.0) * psi[i0] + 1j * cm.sin(
                        Angles[0] / 2.0) *  psi[i3], \
                                                cm.cos(Angles[0] / 2.0) * psi[i3] + 1j * cm.sin(
                                                    Angles[0] / 2.0) *  psi[i0]
                    psi[i1], psi[i2] = cm.cos(Angles[0] / 2.0) *  psi[i1] + (-1j) * cm.sin(
                        Angles[0] / 2.0) *  psi[i2], \
                                                cm.cos(Angles[0] / 2.0) * psi[i2] + (-1j) * cm.sin(
                                                    Angles[0] / 2.0) * psi[i1]

                elif Gate_type == 'rzz':
                    psi[i0], psi[i1],  psi[i2],  psi[i3] = cm.exp(-0.5j * Angles[0]) *  psi[
                        i0], cm.exp(0.5j * Angles[0]) *  psi[i1], \
                                                                                cm.exp(0.5j * Angles[0]) *  psi[
                                                                                    i2], cm.exp(-0.5j * Angles[0]) * \
                                                                                psi[i3]

    return psi



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/7/2 14:37
@Function: test
@Contact: cuijinghao@tgqs.net
"""
from tgqSim.circuit.QuantumCircuit import QuantumCircuit
import tgqSim as tgqs
import numpy as np
from tgqSim.sim.QuantumSimulator import QuantumSimulator


if __name__ == '__main__':
    # use case as below
    nQubits = 4
    qc = QuantumCircuit(0)
    qc.add_qubits(nQubits, name='qft')
    # hgatetest = tgqs.h(1)
    # hgate = tgqs.h(1)
    # # print(hgate)
    # qc.append(hgate)
    # print(qc.gate_list)

    hgate = tgqs.h(0)
    print(hgate)
    qc.append(hgate)
    # print(qc.gate_list)

    # qc.append(tgqs.measure([2,1,3,0]))
    print(qc.gate_list)
    # rxgate = tgqs.rx(qbit=1, theta=0.707)
    # # print(rxgate)
    # qc.append(rxgate)
    # print(qc.gate_list)
    # rxxgate = tgqs.rxx(0,1,0.707)
    # qc.append(rxxgate)
    # print(qc.gate_list)
    # ccxgate = tgqs.ccx([0,2],1)
    # qc.append(ccxgate)
    # print(qc.gate_list)
    simulator = QuantumSimulator()
    # # get state vector
    simulated_state = simulator.run_statevector(qc, 'CPU')
    print(simulated_state)
    #
    # # get probability vector
    result = simulator.execute(qc)
    print(result)
    # qc.add_single_gate(hgatetest)
    # qc.append()
    # for i in range(nQubits):
    #     if nQubits - 2 == i:
    #         qc.x(i)
    #     else:
    #         qc.h(i)
    # for i in range(nQubits - 1, -1, -1):
    #     for j in range(i, -1, -1):
    #         if j == i:
    #             qc.h(j)
    #         else:
    #             qc.cp(control_qubit=j, target_qubit=i, theta=np.pi / (2 ** (i - j)))
    # qc.x(0)
    # for i in range(0, nQubits // 2):
    #     qc.swap(qubit_1=i, qubit_2=nQubits - 1 - i)
    #
    # simulator = QuantumSimulator()
    # simulator.
    # qc.run_statevector()
    # print(qc.state)
    # print(len(qc.state))
    # measure_pos = sorted([1, 0], reverse=True)
    # print(qc.execute(measure_bits_list=measure_pos))
    # qc.show_quantum_circuit()



from __future__ import absolute_import

from tgqSim.circuit.common_gates import CONTROLLED_GATE
from tgqSim.circuit.QuantumCircuit import QuantumCircuit
from tgqSim.sim.QuantumSimulator import QuantumSimulator
from tgqSim.circuit.common_gates import (
    x, y, z,
    h, cx, swap, iswap, cz, cp,
    rx, ry, rz,
    rxx, ryy, rzz,
    s, sdg, t, tdg,
    syc,
    ccx, cswap,
    measure, sqrt_x)

from tgqSim.openqasm2.qasmParser import QASMParser
from tgqSim.openqasm2.qasmGenerator import qasm
from tgqSim.device.noise_util import parse_noise
from tgqSim.device.noise_models import *
from tgqSim.transformers.common_transformers import *
from tgqSim.transformers.common_decompositions import *

__version__ = '1.0.0'
__project__ = 'tgqSim'



import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as lines
from matplotlib.axes._axes import Axes
import numpy as np 

def enumerate_gates(l:list, schedule:bool=False):
    """
    遍历门序列

    两种方式进行遍历：
    一种方式直接遍历列序列，另外一种就是遍历所有的门
    当schedule为True时，则遍历所有的门序列；否则仅遍历列序列

    Args:
        l (list): 门序列，数组中每一个元素存放的是每一列的门序列
        schedule (bool, optional): 默认值是False，决定是按列遍历还是遍历所有门

    Yields:
        当schedule=True，对应的返回类型是：(int, tuple)，分别表示门序列对应的列数以及门信息
        当schedule=False，对应的返回类型是：(int, list)，分别表示门序列对应的列数以及该列所有门序列
    """
    if schedule:
        for i,gates in enumerate(l):
            for gate in gates:
                yield i,gate
    else:
        for i,gate in enumerate(l):
            yield i,gate
    return

def measured_wires(l:list, labels:list, schedule:bool=False)->dict:
    """
    确定测量门在那个比特位上，主要是确定测量门具体位置，方便后面绘制测量门

    Args:
        l (list): 门序列，数组中每一个元素存放的是每一列的门序列
        labels (list): 里面元素都是字符串，列举出每一个比特的标签
        schedule (bool, optional): 默认值是False，主要决定是按照所有门进行遍历还是按列遍历.

    Returns:
        dict: 获取一个映射关系，key是比特索引值，value是对应的列数，可以确定测量位置在哪里，方便绘图
    """
    measured = {}
    for i,gate in enumerate_gates(l, schedule=schedule):
        name,target = gate[:2]
        if isinstance(target, tuple):
            continue
        target = f"q_{target}"
        # 其主要是目的是更新测量们的起始索引，这个是方式一行有多个测量门，仅以最后一个测量门为起始索引
        # j是对应的比特位
        j = get_flipped_index(target, labels)
        if name.startswith('M'):
            measured[j] = i
    return measured

def draw_gates(ax:Axes, l:list, labels:list, gate_grid_index:list, wire_grid:np.array, plot_params:dict, measured:dict={}, schedule:bool=False):
    """
    绘制控制门
    具体逻辑如下：
    首先判断当前门的信息长度是否超过2，若是超过两个，则存在控制门（RXX、RYY、RZZ、ISWAP等都是特殊处理），则先绘制目标门，然后再绘制控制门
    若是长度不超过2，则是单门，只需要绘制目标门即可

    Args:
        ax (Axes): 绘图对象
        l (list): 门序列，数组中每一个元素存放的是每一列的门序列
        labels (list): 里面元素都是字符串，列举出每一个比特的标签
        gate_grid_index (list): 每一个比特位，对应的起始索引值（横向）
        wire_grid (np.array): 纵向位置
        plot_params (dict): 基本参数信息
        measured (dict, optional): 测量位置信息
        schedule (bool, optional): 默认值是False，主要决定是按照所有门进行遍历还是按列遍历.
    """
    for i,gate in enumerate_gates(l, schedule=schedule):
        # print(gate)
        if len(gate) > 2: # Controlled
            gate_grid_index = draw_with_controls(ax, gate, labels, gate_grid_index, wire_grid, plot_params, measured)
        else:
            if isinstance(gate[-1], tuple):
                x = max(gate_grid_index[gate[-1][0]], gate_grid_index[gate[-1][1]])
                min_index, max_index = min(gate[-1]), max(gate[-1])
                for i in range(min_index, max_index + 1):
                    gate_grid_index[i] = x + plot_params["scale"]
            else:
                target = gate[-1]
                x = gate_grid_index[target]
                gate_grid_index[target] += plot_params["scale"]
            draw_target(ax, x, gate, labels, wire_grid, plot_params)
    return

def draw_with_controls(ax:Axes, gate:tuple, labels:list, gate_grid_index:list, wire_grid:np.array, plot_params:dict, measured:dict={}):
    """
    绘制带有控制位的门

    Args:
        ax (Axes): 绘图对象
        gate (tuple): 门信息，带有门名字、目标位置以及控制位置
        labels (list): 里面元素都是字符串，列举出每一个比特的标签
        gate_grid_index (list): 每一个比特位，对应的起始索引值（横向）
        wire_grid (np.array): 纵向位置
        plot_params (dict): 基本参数信息
        measured (dict, optional): 测量位置信息

    Returns:
        list: 跟新后的每一个比特位对应的起始索引值（横向）
    """
    scale = plot_params['scale']
    name,targets = gate[:2]
    target = f"q_{targets}"
    target_index = get_flipped_index(target, labels)
    controls_index = gate[2:]
    controls = [f"q_{t}" for t in controls_index]
    control_indices = get_flipped_indices(controls, labels)
    gate_indices = control_indices + [target_index]
    min_wire = min(gate_indices)
    max_wire = max(gate_indices)
    max_col = 0
    target_contrl_index = list(controls_index) + [targets]
    min_index, max_index = min(target_contrl_index), max(target_contrl_index)
    for p in range(min_index, max_index + 1):
        if gate_grid_index[p] > max_col:
            max_col = gate_grid_index[p]
    for p in range(min_index, max_index + 1):
        gate_grid_index[p] = max_col + scale
    line(ax, max_col, max_col, wire_grid[min_wire], wire_grid[max_wire], plot_params, '#7A56D0')
    draw_target(ax, max_col, gate, labels, wire_grid, plot_params)

    for ci in control_indices:
        x = max_col
        y = wire_grid[ci]
        if name in ['SWAP']:
            swapx(ax,x,y,plot_params)
        else:
            cdot(ax,x,y,plot_params)
    return gate_grid_index

def draw_target(ax:Axes, x:float, gate:tuple, labels:list, wire_grid:np.array, plot_params:dict):
    """
    绘制目标位信息

    Args:
        ax (Axes): 绘图对象
        x (float): 横坐标位置信息
        gate (tuple): 门信息，带有门名字、目标位置以及控制位置
        labels (list): 里面元素都是字符串，列举出每一个比特的标签
        wire_grid (np.array): 纵向位置
        plot_params (dict): 基本参数信息
    """
    rectangle_delta = plot_params['rectangle_delta']
    name,target = gate[:2]
    if isinstance(target, tuple):
        # print(target)
        if 2 == len(target):
            target1 = f"q_{target[0]}"
            target2 = f'q_{target[1]}'
            target_index1 = get_flipped_index(target1, labels)
            target_index2 = get_flipped_index(target2, labels)
            y1 = wire_grid[target_index1]
            y2 = wire_grid[target_index2]
            if name in ["RXX", 'RYY', 'RZZ']:
                textstr = r'$%s_{%s}$' % (name[0], name[1:])
            elif name == 'ISWAP':
                textstr = 'iSWAP'
            else:
                textstr = name
            rectangle(ax, x - rectangle_delta, x + rectangle_delta, y1, y2, textstr, plot_params)
            return
    target = f"q_{target}"
    # x = gate_grid[target]
    target_index = get_flipped_index(target, labels)
    y = wire_grid[target_index]
    if name in ['CX','CCX']:
        oplus(ax, x, y, plot_params)
    elif name in ['CZ']:
        cdot(ax, x, y, plot_params)
    elif name in ['SWAP']:
        swapx(ax, x, y, plot_params)
    elif name in ["RX", "RY", "RZ"]:
        rotate_axis(ax, x, y, name, plot_params, box=True)
    else:
        text(ax, x, y, name, plot_params, box=True)
    return

def rectangle(ax:Axes, x1:float, x2:float, y1:float, y2:float, textstr:str, plot_params:dict):
    """
    绘制RXX, RYY, RZZ, ISWAP等用矩形来表示的门

    Args:
        ax (Axes): 绘图对象
        x1 (float): 矩形左下角横坐标
        x2 (float): 矩形右上角横坐标
        y1 (float): 矩形左下角纵坐标
        y2 (float): 矩形右上角纵坐标
        textstr (str): 在矩形中需要写入的内容
        plot_params (dict): 基本参数信息
    """
    rectangle_delta = plot_params['rectangle_delta']
    if x1 > x2:
        x1, x2 = x2, x1
    if y1 > y2:
        y1, y2 = y2, y1
    fontsize = plot_params['fontsize']
    linewidth = plot_params['linewidth']
    y1_new = y1 - rectangle_delta
    y2_new = y2 + rectangle_delta
    width = x2 - x1
    height = y2_new - y1_new
    # rectangle = Rectangle((x1, y1_new), width, height, fc='w', ec='k', alpha=1.0, zorder=2)
    rectangle = patches.FancyBboxPatch((x1, y1_new), width, height, boxstyle='round,pad=0.09', ec='#7A56D0', fc='#E1BEE7', zorder=2, lw=linewidth)
    ax.add_patch(rectangle)
    # 开始画出对应的作用的比特
    Circle = patches.Circle
    control_radius = plot_params['control_radius']
    c = Circle((x1, y1), control_radius, ec='#7A56D0',
               fc='#7A56D0', fill=True, lw=linewidth, zorder=2)
    ax.add_patch(c)
    ax.text(x1 + rectangle_delta / 2, y1, '$1$', color='#7A56D0', ha='center', va='center', size=fontsize)
    c = Circle((x1, y2), control_radius, ec='#7A56D0',
               fc='#7A56D0', fill=True, lw=linewidth, zorder=2)
    ax.add_patch(c)
    ax.text(x1 + rectangle_delta / 2, y2, '$0$', color='#7A56D0', ha='center', va='center', size=fontsize)
    if 'iSWAP' == textstr:
        fontsize = 11.0
    ax.text(x1 + rectangle_delta, y1_new + height / 2, textstr, color='#7A56D0', ha='center', va='center', size=fontsize)
    return 

def line(ax:Axes, x1:float, x2:float, y1:float, y2:float, plot_params:dict, linecolor:str='#50C1E9'):
    """
    绘制直线

    Args:
        ax (Axes): 绘图对象
        x1 (float): 线段第一个点的横坐标
        x2 (float): 线段第二个点的横坐标
        y1 (float): 线段第一个点的纵坐标
        y2 (float): 线段第二个点的纵坐标
        plot_params (dict): 基本参数信息
        linecolor (str, optional): 线段的颜色，默认值是'#50C1E9'
    """
    Line2D = lines.Line2D
    line = Line2D((x1, x2), (y1, y2),
        color=linecolor,lw=plot_params['linewidth'])
    ax.add_line(line)

def rotate_axis(ax:Axes, x:float, y:float, textstr:str, plot_params:dict, box:bool=False):
    """
    主要是绘制RX, RY, RZ
    Args:
        ax (Axes): 绘图对象
        x (float): box中写入文字的横坐标
        y (float): box中写入文字的纵坐标
        textstr (str): 在box中需要写入的内容
        plot_params (dict): 基本参数信息
        box (bool, optional): 主要是决定是否需要重新设置bbox，默认值是False
    """
    linewidth = plot_params['linewidth']
    fontsize = plot_params['fontsize']
    box_pad = plot_params['box_pad']
    if box:
        bbox = dict(ec='#50C1E9', fc='#50C1E9', fill=True, lw=linewidth, boxstyle='round,pad=0.5')
    else:
        bbox= dict(fill=False,lw=0)
    textstr = r"$%s_%s$" % (textstr[0], textstr[1])
    targetbox = patches.FancyBboxPatch(xy=(x - box_pad, y - box_pad), height=2 * box_pad, width=2 * box_pad, ec='#50C1E9', fc='#50C1E9', fill=True, lw=linewidth, boxstyle='round,pad=0.08', zorder=2, alpha=1.0)
    ax.add_patch(targetbox)
    ax.text(x, y,textstr,color='w',ha='center',va='center',bbox=bbox,size=fontsize)
    return

def text(ax:Axes, x:float, y:float, textstr:str, plot_params:dict, textcolor:str="w", box:bool=False):
    """
    绘制一般的单门，如H, X, Y, Z等
    Args:
        ax (Axes): 绘图对象
        x (float): box中写入文字的横坐标
        y (float): box中写入文字的纵坐标
        textstr (str): 在box中需要写入的内容
        plot_params (dict): 基本参数信息
        textcolor (str, optional): 默认值是w，对应的文字颜色
        box (bool, optional): 主要是决定是否需要重新设置bbox，默认值是False
    """
    linewidth = plot_params['linewidth']
    fontsize = plot_params['fontsize']
    box_pad = plot_params['box_pad']
    # if box:
    #     bbox = dict(ec='#50C1E9', fc='#50C1E9', fill=True, lw=linewidth, boxstyle='round,pad=0.5')
    # else:
    #     bbox= dict(fill=False,lw=0)
    if textstr.endswith('DG'):
        textstr = r'$%s^\dagger$' % textstr[0]
    targetbox = patches.FancyBboxPatch(xy=(x - box_pad, y - box_pad), height=2 * box_pad, width=2 * box_pad, ec='#50C1E9', fc='#50C1E9', fill=True, lw=linewidth, boxstyle='round,pad=0.08', zorder=2)
    ax.add_patch(targetbox)
    ax.text(x, y, textstr, color=textcolor, ha='center', va='center', size=fontsize)
    return

def oplus(ax:Axes, x:float, y:float, plot_params:dict):
    """
    绘制CX和CCX这两个门，主要是解决直和那个符号
    Args:
        ax (Axes): 绘图对象
        x (float): 对应的圆心横坐标
        y (float): 对应的圆心纵坐标
        plot_params (dict): 基本参数信息
    """
    Circle = patches.Circle
    not_radius = plot_params['not_radius']
    linewidth = plot_params['linewidth']
    c = Circle((x, y), not_radius, ec='#7A56D0',
               fc='#7A56D0', fill=False, lw=linewidth)
    ax.add_patch(c)
    line(ax, x, x, y - not_radius, y + not_radius, plot_params, '#7A56D0')
    line(ax, x - not_radius, x + not_radius, y, y, plot_params, '#7A56D0')
    return

def cdot(ax:Axes, x:float, y:float, plot_params:dict, dotcolor:str='#7A56D0'):
    """
    绘制控制位那个点
    Args:
        ax (Axes): 绘图对象
        x (float): 对应的圆心横坐标
        y (float): 对应的圆心纵坐标
        plot_params (dict): 基本参数信息
        dotcolor (str, optional): 设置点的颜色，默认值是'#7A56D0'.
    """
    Circle = patches.Circle
    control_radius = plot_params['control_radius']
    scale = plot_params['scale']
    linewidth = plot_params['linewidth']
    c = Circle((x, y), control_radius*scale,
        ec=dotcolor, fc=dotcolor, fill=True, lw=linewidth, zorder=2)
    ax.add_patch(c)
    return

def swapx(ax:Axes, x:float, y:float, plot_params:dict):
    """
    绘制交换门
    Args:
        ax (Axes): 绘图对象
        x (float): 绘制❌中心位置的横坐标
        y (float): 绘制❌中心位置的纵坐标
        plot_params (dict): 基本参数信息
    """
    d = plot_params['swap_delta']
    line(ax, x-d, x+d, y-d, y+d, plot_params, linecolor='#7A56D0')
    line(ax, x-d, x+d, y+d, y-d, plot_params, linecolor='#7A56D0')
    return

def setup_figure(nq:int, ng:int, gate_grid:np.array, wire_grid:np.array, plot_params:dict):
    """
    设置图形的基本信息
    Args:
        nq (int): 比特数
        ng (int): 门序列列数，主要是设置整个图的宽度
        gate_grid (np.array): 整个图的所有点的横坐标
        wire_grid (np.array): 整个图的所有点的纵坐标
        plot_params (dict): 基本参数信息
    """
    scale = plot_params['scale']
    fig = plt.figure(
        figsize=(ng * scale, nq * scale),
        facecolor='w',
        edgecolor='w'
    )
    ax = fig.add_subplot(1, 1, 1,frameon=True)
    # fig.tight_layout()
    ax.set_axis_off()
    offset = 0.5*scale
    ax.set_xlim(gate_grid[0] - offset, gate_grid[-1] + offset)
    ax.set_ylim(wire_grid[0] - offset, wire_grid[-1] + offset)
    ax.set_aspect('equal')
    return fig,ax

def draw_wires(ax:Axes, nq:int, gate_grid: np.array, wire_grid:np.array, plot_params:dict, linecolor:str, measured:dict={}):
    """
    绘制线路图中横线
    Args:
        ax (Axes): 绘图对象
        nq (int): 比特数
        gate_grid (np.array): 整个图的所有点的横坐标
        wire_grid (np.array): 整个图的所有点的纵坐标
        plot_params (dict): 基本参数信息
        linecolor (str): 线段的颜色
        measured (dict, optional): 测量门位置信息
    """
    scale = plot_params['scale']
    for i in range(nq):
        line(ax, gate_grid[0] - scale, gate_grid[-1] + scale, wire_grid[i], wire_grid[i], plot_params, linecolor)
        
    return

def draw_labels(ax:Axes, labels:list, inits:dict, gate_grid:list, wire_grid:np.array, plot_params:dict, textcolor:str):
    """
    每根线谱前对应的标签
    Args:
        ax (Axes): 绘图对象
        labels (list): 里面元素都是字符串，列举出每一个比特的标签
        inits (dict): 线路图中每一条线前面的标签与索引的关系
        gate_grid (list): 整个图的所有点的横坐标
        wire_grid (np.array): 整个图的所有点的纵坐标
        plot_params (dict): 基本参数信息
        textcolor (str): 填写内容的颜色
    """
    scale = plot_params['scale']
    label_buffer = plot_params['label_buffer']
    nq = len(labels)
    for i in range(nq):
        j = get_flipped_index(labels[i], labels)
        text(ax, gate_grid[0] - label_buffer * scale, wire_grid[j], render_label(labels[i], inits), plot_params, textcolor)
    return

def get_flipped_index(target:str, labels:list)->int:
    """
    获取当前门对应的比特位置索引
    Args:
        target (str): 门的名称
        labels (list): 里面元素都是字符串，列举出每一个比特的标签

    Returns:
        int: 对应整个线路图中纵坐标对应的索引值
    """
    nq = len(labels)
    i = labels.index(target)
    # nq-i-1是因为plt画图是：从下到上逐渐增大，但是比特位需要是从上到下是逐渐增大
    return nq-i-1

def get_flipped_indices(targets:list, labels:list)->list: 
    """
    获取多个门位置的比特位索引值
    Args:
        targets (list): 门序列，主要是用于处理多个控制位
        labels (list): 里面元素都是字符串，列举出每一个比特的标签

    Returns:
        list: 门序列中对应的比特位索引
    """
    return [get_flipped_index(t, labels) for t in targets]

def render_label(label:str, inits:dict={}):
    """
    绘制前面标签内容
    Args:
        label (str): 在当前代码中比特位的标签
        inits (dict, optional): 线路图中每一条线前面的标签与索引的关系

    Returns:
        str: Latex字符串代码
    """
    if label in inits:
        s = inits[label]
        if s is None:
            return ''
        else:
            return r'$q_%s|0\rangle$' % inits[label]
    return r'$q_%s|0\rangle$' % label



"""
-*- coding: utf-8 -*-
@Author : Cui Jinghao
@Time : 2024/6/28 14:43
@Function: noise_model.py
@Contact: cuijinghao@tgqs.net
"""


import random
from typing import Union

NOISE_MAPPER = {"bit_flip": "BF",
                "asymmetric_depolarization": "ADP",
                "depolarize": "DP",
                "phase_flip": "PF",
                "phase_damp": "PD",
                "amplitude_damp": "AD",
                "damp_I": "DAMP_I"}

NOISE_TYPE = ["bit_flip",
              # "asymmetric_depolarization",
              "depolarize", "phase_flip", "phase_damp", "amplitude_damp"]


def bit_flip(error_rate: float):
    random_value = random.uniform(0, 1)
    gate_name = "I"
    if random_value < error_rate:
        gate_name = "x"
    return gate_name


def asymmetric_depolarization(error_rate: Union[list, tuple]):
    if len(error_rate) != 3:
        raise ValueError("Length of error must be 3")
    gate_name = "I"
    thredhold = sum(error_rate)
    random_value = random.uniform(0, 1)
    if random_value < thredhold:
        if random_value < error_rate[0]:
            gate_name = "x"
        elif random_value < sum(error_rate[:2]):
            gate_name = "y"
        else:
            gate_name = "z"
    return gate_name


def depolarize(error_rate: float):
    pxyz = [error_rate / 3] * 3
    return asymmetric_depolarization(error_rate=pxyz)


def phase_flip(error_rate: float):
    random_value = random.uniform(0, 1)
    gate_name = "I"
    if random_value < error_rate:
        gate_name = "z"
    return gate_name


def phase_damp(error_rate: float):
    random_value = random.uniform(0, 1)
    gate_name = "damp_I"
    if random_value < error_rate:
        gate_name = "pd"
    return gate_name


def amplitude_damp(error_rate: float):
    random_value = random.uniform(0, 1)
    gate_name = "damp_I"
    if random_value < error_rate:
        gate_name = "ad"
    return gate_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author: Yuchen He
@contact: heyuchen@tgqs.net
@version: 1.0.1
@file: QuantumCircuit.py
@time: 2024/1/16 17:16

@Modifider: Zhiqiang Wang
@contact: wangzhiqiang@tgqs.net
@version: 0.1.1
@file: QuantumCircuit.py
@time: 2024/04/19 15:36
"""
from .tgqSim.CpuGate import SingleGate, DoubleGate, TripleGate
from .tgqSim.utils import dev_tools
from typing import Union
import random, os, ctypes
import numpy as np
import tgqSim.draw_circuit_tools as tools
import matplotlib.pyplot as plt
import GPUtil, datetime
# from numba import njit, prange

class Float2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float),
                ("y", ctypes.c_float)]

class Double2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_double),
                ("y", ctypes.c_double)]

class GateInfo(ctypes.Structure):
    _fields_ = [
        ("gateName", ctypes.c_char_p),
        ("actionPos", ctypes.POINTER(ctypes.c_int)),
        ("theta", ctypes.c_double)
    ]


class Qubit:
    def __init__(self, name=None, idx=None, pos=None):
        self.label = 'p' if not name else name
        self.name = f'p_{idx+100}' if not name else f"{name}_{idx}"
        self.pos = pos + idx
        self.idx = idx

    # def __getattr__(self, item):
    #     if item == self.name:
    #         return self.idx
    #     if item == self.label+"["+str(self.pos)+"]":
    #         return self.idx

    def __str__(self):
        return str(self.name)


class SimulationError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class QuantumCircuit:
    def __init__(self, qbit_number=0):
        self.qbit_list = []
        self.cbit_list = []
        self.gate_list = []
        self.qbit_register = {}
        self.base_single_gate = ['h', 'x', 'y', 'z', 'rx', 'ry', 'rz', 'u3', 's', 'sdg', 't', 'tdg', "damp_I", "pd", "ad"]
        self.base_double_gate = ['cx', 'swap', 'iswap', 'cz', 'cp', 'rxx', 'ryy', 'rzz', 'syc']
        self.base_triple_gate = ['ccx', 'cswap']
        self.noise_mapper = {"bit_flip": "BF", "asymmetric_depolarization": "ADP", "depolarize": "DP", "phase_flip": "PF", "phase_damp": "PD", "amplitude_damp": "AD"}
        self.width = qbit_number
        self.state = []
        self.circuit_diag = []
        self.isgpu = False
        self.deviceid = []
        self.noise_circuit = []
        self.current_gate_info = ()
        self.prob_result = {}

    def add_qubits(self, number: int, name: str = 'p'):
        """
        添加量子比特
        :param number: 比特数量
        :param name: 比特名称
        :return:
        """
        self.width += number
        if name not in self.qbit_register:
            self.qbit_register.update({name: number})
            sub_list = [Qubit(name=name, idx=i, pos=i+self.width) for i in range(number)]
        else:
            former_number = self.qbit_register[name]
            self.qbit_register.update({name: number+former_number})
            sub_list = [Qubit(name=name, idx=i+former_number, pos=i+self.width) for i in range(number)]
        self.qbit_list.append(sub_list)

    def add_single_gate(self, gate_pos: Union[int, str], *gate_info):
        """
        添加单比特门序列
        :param gate_pos:
        :param gate_info: 门类型，门参数
        :return:
        """
        tmp_pos = gate_pos
        if isinstance(gate_pos, list):
            tmp_pos = gate_pos[0]
        if tmp_pos < 0 or tmp_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in self.base_single_gate:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        if isinstance(gate_pos, int):
            gatename = gate_info[0].upper()
            self.circuit_diag.append([(gatename, gate_pos)])
        else:
            self.circuit_diag.append([(gatename, int(gate_pos))])

        self.current_gate_info = (gate_pos, gate_info)
        self.noise_circuit.append((gate_pos, gate_info))
        self.gate_list.append((gate_pos, gate_info))
        # print("over")
        return self
        # todo 通过比特名称访问
        # if isinstance(gate_pos, str):
        #     return self.gate_list.append((gate_pos, gate_type))

    def add_double_gate(self, gate_pos_0:int, gate_pos_1:int, *gate_info):
        """
        添加两比特门序列
        :param gate_pos_0: 控制位
        :param gate_pos_1: 目标位
        :param gate_info: 门类型，门参数
        :return:
        """
        min_pos, max_pos = min([gate_pos_0, gate_pos_1]), max([gate_pos_0, gate_pos_1])
        if min_pos < 0 or max_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in self.base_double_gate:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        if gate_info[0] in ["rxx", 'ryy', 'rzz', 'iswap', 'syc']:
            self.circuit_diag.append([(gate_info[0].upper(), (gate_pos_1, gate_pos_0))])
        else:
            self.circuit_diag.append([(gate_info[0].upper(), gate_pos_1, gate_pos_0)])
        self.current_gate_info = ([gate_pos_0, gate_pos_1], gate_info)
        self.noise_circuit.append(([gate_pos_0, gate_pos_1], gate_info))
        self.gate_list.append(([gate_pos_0, gate_pos_1], gate_info))
        return self

    def add_triple_gate(self, gate_pos:list, *gate_info):
        """
        添加三比特门序列
        :param gate_pos:
        :param gate_info:
        :return:
        """
        min_pos, max_pos = min(gate_pos), max(gate_pos)
        if min_pos < 0 or max_pos >= self.width:
            raise ValueError("添加的位置不能为负数或不能超过线路设置的比特数")
        if len(gate_info) >= 1:
            if gate_info[0] not in self.base_triple_gate:
                raise ValueError("添加的门名称在框架支持中，请检查添加门的名称")
        self.circuit_diag.append([(gate_info[0].upper(), gate_pos[2], gate_pos[0], gate_pos[1])])
        self.current_gate_info = (gate_pos, gate_info)
        self.noise_circuit.append((gate_pos, gate_info))
        self.gate_list.append((gate_pos, gate_info))
        return self
    
    def setdevice(self, deviceList: Union[int, list]):
        gpus = GPUtil.getGPUs()
        gpuidList = [gpu.id for gpu in gpus]
        if isinstance(deviceList, int):
            deviceList = [deviceList]
        for deviceid in deviceList:
            if deviceid not in gpuidList:
                raise ValueError("设置设备ID不存在")
        self.isgpu = True
        self.deviceid = deviceList

    @staticmethod
    def get_cuda_lib():
        cuda_version = dev_tools.get_cuda_version().replace(".", "-")
        lib_name = f"cuda_{cuda_version}_tgq_simulator.so"
        current_file_path = os.path.abspath(__file__)
        current_directory = os.path.dirname(current_file_path)
        dll_path = os.path.abspath(current_directory + '/lib/' + lib_name)
        lib = ctypes.CDLL(dll_path)
        return lib

    def _run_with_device(self):
        lib = QuantumCircuit.get_cuda_lib()
        lib.execute_circuit.argtypes = [
            ctypes.POINTER(ctypes.POINTER(Double2)),
            ctypes.POINTER(GateInfo),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int)
        ]
        lib.execute_circuit.restype = None
        gateInfo = []
        for (gate_pos, gate_info) in self.gate_list:
            if isinstance(gate_pos, int):
                length = 2
                gate_pos = [gate_pos]
            elif isinstance(gate_pos, list):
                length = len(gate_pos) + 1
            else:
                raise TypeError("Type of gate_pos must be int or list")
            gate_obj = GateInfo()
            actionPos = gate_pos + [-1]
            gate_obj.actionPos = (ctypes.c_int * length)(*actionPos)
            if len(gate_info) > 0:
                gate_obj.gateName = gate_info[0].encode(encoding='utf-8')
            if len(gate_info) > 1:
                gate_obj.theta = gate_info[1]
            gateInfo.append(gate_obj)
        gateInfoCData = (GateInfo * len(gateInfo))(*gateInfo)
        deviceIdCData = (ctypes.c_int * len(self.deviceid))(*self.deviceid)
        # 申请内存首地址，不在Python端申请内存
        # 在C语言中申请统一内存，减少多次拷贝动作
        self.state = ctypes.POINTER(Double2)()
        lib.execute_circuit(ctypes.byref(self.state), gateInfoCData, len(gateInfo), self.width, deviceIdCData)
        # iStart = datetime.datetime.now()
        # print(f"start time is {iStart}")
        self.state = np.ctypeslib.as_array(self.state, shape=(2**self.width,))
        self.state = self.state.view(np.complex128)
        # print(f"total time of changing type is {(datetime.datetime.now() - iStart).total_seconds()} secs")
        
    # todo:
    def apply_state_space(self):
        self.state = ctypes.POINTER(Double2)()
        return self.state

    def free_state(self):
        if self.isgpu:
            lib = QuantumCircuit.get_cuda_lib()
            lib.freeAllMem.argtypes = [
                np.ctypeslib.ndpointer(dtype=np.complex128)
            ]
            lib.freeAllMem.restype = None
            lib.freeAllMem(self.state)

    def run_with_noise(self, shots:int=1000):
        noise_type = ["bit_flip", "asymmetric_depolarization", "depolarize", "phase_flip", "phase_damp", "amplitude_damp"]
        result_dict = {}
        tmp_circuit = self.gate_list
        for _ in range(shots):
            new_circuit = []
            for (gate_pos, gate_info) in self.noise_circuit:
                if gate_info[0] in noise_type:
                    noise_gate = QuantumCircuit.parse_noise(noise_type=gate_info[0], gate_pos=gate_pos, error_rate=gate_info[1])
                    # print(noise_gate)
                    if noise_gate is not None:
                        new_circuit.append(noise_gate)
                else:
                    new_circuit.append((gate_pos, gate_info))
            # print("new_circuit:", new_circuit)
            self.gate_list = new_circuit
            result = self.measure(measure_bits_list=[i for i in range(self.width)], shots=1000)

            # print(self.state)
            for key in result.keys():
                if key in result_dict:
                    result_dict[key] += result[key]
                else:
                    result_dict[key] = result[key]
        self.gate_list = tmp_circuit
        self.prob_result = dev_tools.get_normalization(frequency=result_dict)

    def run_statevector(self):
        """
        根据线路的门序列计算末态的量子态
        :return:
        """            
        if not self.isgpu:
            self.state = [1 if a == 0 else 0 for a in range(2**self.width)]
            for (gate_pos, gate_info) in self.gate_list:
                gate_type = gate_info[0]
                angle = tuple(gate_info[1:])
                if gate_type in self.base_single_gate:
                    self.state = SingleGate.ActOn_State(self.state, self.width, gate_type, gate_pos, *angle)

                elif gate_type in self.base_double_gate:
                    set_gate_pos = set(gate_pos)
                    if len(set_gate_pos) != len(gate_pos):
                        raise SimulationError(f"Gate position cannot be the same: {gate_pos[0]}, {gate_pos[1]}")
                    self.state = DoubleGate.ActOn_State(self.state, self.width, gate_type, gate_pos, *angle)
                elif gate_type in self.base_triple_gate:
                    set_gate_pos = set(gate_pos)
                    if len(set_gate_pos) != len(gate_pos):
                        raise SimulationError(f"Gate position cannot be the same: "
                                            f"{gate_pos[0]}, {gate_pos[1]} and {gate_pos[2]}")
                    self.state = TripleGate.ActOn_State(self.state, self.width, gate_type, gate_pos, *angle)
                else:
                    raise SimulationError(f"Unkown gate type: {gate_type}")
        else:
            self._run_with_device()
        
    def with_noise(self, noise_type:str, error_rate:Union[float, list]):
        """
        将当前的所有控制门都加上noise_type类型的噪声

        Args:
            noise_type (str): 噪声类型
            error_rate (Union[float, list]): 错误率
        """
        # self.noise_circuit += [self.current_gate_info]
        gate_pos = self.current_gate_info[0]
        if isinstance(gate_pos, int):
            self.noise_circuit += [(gate_pos, (noise_type, error_rate))]
            self.circuit_diag.append([(self.noise_mapper[noise_type], gate_pos)])
        else:
            for gate in gate_pos:
                self.noise_circuit += [(gate, (noise_type, error_rate))]
                self.circuit_diag.append([(self.noise_mapper[noise_type], gate)])
    
    @staticmethod
    def parse_noise(noise_type: str,  gate_pos: int, error_rate: Union[float, list]):
        """
        对于不同的噪声类型，返回不同的信息

        Args:
            noise_type (str): 噪声类型
            error_rate Union[float, list]: 门错误率信息
        """
        gate_name = "I"
        if "bit_flip" == noise_type:
            gate_name = QuantumCircuit.bit_flip(error_rate)
        elif "asymmetric_depolarization" == noise_type:
            gate_name = QuantumCircuit.asymmetric_depolarization(error_rate)
        elif "depolarize" == noise_type:
            gate_name = QuantumCircuit.depolarize(error_rate)
        elif "phase_flip" == noise_type:
            gate_name = QuantumCircuit.phase_flip(error_rate)
        elif "phase_damp" == noise_type:
            gate_name = QuantumCircuit.phase_damp(error_rate)
        elif "amplitude_damp" == noise_type:
            gate_name = QuantumCircuit.amplitude_damp(error_rate)
        else:
            raise ValueError("Please check noise model name")
        if gate_name == "I":
            return None
        return (gate_pos, (gate_name,))
    
    @staticmethod
    def bit_flip(error_rate: float):
        random_value = random.uniform(0, 1)
        gate_name = "I"
        if random_value < error_rate:
            gate_name = "x"
        return gate_name
    
    @staticmethod
    def asymmetric_depolarization(error_rate: Union[list, tuple]):
        if len(error_rate) != 3:
            raise ValueError("Length of error must be 3")
        gate_name = "I"
        thredhold = sum(error_rate)
        random_value = random.uniform(0, 1)
        if random_value < thredhold:
            if random_value < error_rate[0]:
                gate_name = "x"
            elif random_value < sum(error_rate[:2]):
                gate_name = "y"
            else:
                gate_name = "z"
        return gate_name
    
    @staticmethod
    def depolarize(error_rate: float):
        pxyz = [error_rate / 3] * 3
        return QuantumCircuit.asymmetric_depolarization(error_rate=pxyz)

    @staticmethod
    def phase_flip(error_rate: float):
        random_value = random.uniform(0, 1)
        gate_name = "I"
        if random_value < error_rate:
            gate_name = "z"
        return gate_name

    @staticmethod
    def phase_damp(error_rate: float):
        random_value = random.uniform(0, 1)
        gate_name = "damp_I"
        if random_value < error_rate:
            gate_name = "pd"
        return gate_name

    @staticmethod
    def amplitude_damp(error_rate: float):
        random_value = random.uniform(0, 1)
        gate_name = "damp_I"
        if random_value < error_rate:
            gate_name = "ad"
        return gate_name
    
    def random_circuit(self, num_qubits, num_gates: Union[int, list]):
        """
        生成随机线路
        :param num_qubits:
        :param num_gates: 门数量，列表表示单比特，多比特门数量
        :return:
        """
        self.width = num_qubits
        if isinstance(num_gates, int):
            base_gate_list = self.base_single_gate + self.base_double_gate + self.base_triple_gate
            for _ in range(num_gates):
                gate_type = np.random.choice(base_gate_list)
                angle = np.random.uniform(-2*np.pi, 2*np.pi, size=3)
                if gate_type in self.base_single_gate:
                    gate_pos = np.random.choice(np.array(range(num_qubits)), size=1)[0]
                elif gate_type in self.base_double_gate:
                    gate_pos = random.sample(range(num_qubits), 2)
                    gate_pos = list(gate_pos)
                elif gate_type in self.base_triple_gate:
                    gate_pos = random.sample(range(num_qubits), 3)
                    gate_pos = list(gate_pos)
                else:
                    gate_pos = None
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
        elif isinstance(num_gates, list):
            single_gate_num = num_gates[0]
            multiple_gate_num = num_gates[1]
            for _ in range(single_gate_num):
                gate_type = np.random.choice(self.base_single_gate)
                angle = np.random.uniform(-2*np.pi, 2*np.pi, size=3)
                gate_pos = np.random.choice(np.array(range(num_qubits)), size=1)[0]
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
            for _ in range(multiple_gate_num):
                base_gate_list = self.base_double_gate+self.base_triple_gate
                gate_type = np.random.choice(base_gate_list)
                angle = np.random.uniform(-2 * np.pi, 2 * np.pi, size=3)
                if gate_type in self.base_double_gate:
                    gate_pos = random.sample(range(num_qubits), 2)
                    gate_pos = list(gate_pos)
                elif gate_type in self.base_triple_gate:
                    gate_pos = random.sample(range(num_qubits), 3)
                    gate_pos = list(gate_pos)
                else:
                    gate_pos = None
                self.gate_list.append([gate_pos, tuple([gate_type]) + tuple(angle)])
                random.shuffle(self.gate_list)

    def h(self, qbit: Union[int, str]):
        return self.add_single_gate(qbit, 'h')

    def x(self, qbit):
        return self.add_single_gate(qbit, 'x')

    def y(self, qbit):
        return self.add_single_gate(qbit, 'y')

    def z(self, qbit):
        return self.add_single_gate(qbit, 'z')

    def rx(self, qbit, theta):
        return self.add_single_gate(qbit, 'rx', theta)

    def ry(self, qbit, theta):
        return self.add_single_gate(qbit, 'ry', theta)

    def rz(self, qbit: Union[int, str], theta):
        return self.add_single_gate(qbit, 'rz', theta)

    def u3(self, qbit, *theta):
        return self.add_single_gate(qbit, 'u3', *theta)
    
    def s(self, qbit:Union[int, str]):
        return self.add_single_gate(qbit, 's')
    
    def sdg(self, qbit:Union[int, str]):
        return self.add_single_gate(qbit, 'sdg')
    
    def t(self, qbit:Union[int, str]):
        return self.add_single_gate(qbit, 't')
    
    def tdg(self, qbit:Union[int, str]):
        return self.add_single_gate(qbit, 'tdg')

    def cx(self, control_qubit, target_qubit):
        return self.add_double_gate(control_qubit, target_qubit, 'cx')

    def swap(self, qubit_1, qubit_2):
        return self.add_double_gate(qubit_1, qubit_2, 'swap')

    def iswap(self, qubit_1, qubit_2):
        return self.add_double_gate(qubit_1, qubit_2, 'iswap')

    def cz(self, control_qubit, target_qubit):
        return self.add_double_gate(control_qubit, target_qubit, 'cz')

    def cp(self, control_qubit, target_qubit, theta):
        return self.add_double_gate(control_qubit, target_qubit, 'cp', theta)

    def rxx(self, qubit_1, qubit_2, theta):
        return self.add_double_gate(qubit_1, qubit_2, 'rxx', theta)

    def ryy(self, qubit_1, qubit_2, theta):
        return self.add_double_gate(qubit_1, qubit_2, 'ryy', theta)

    def rzz(self, qubit_1, qubit_2, theta):
        return self.add_double_gate(qubit_1, qubit_2, 'rzz', theta)

    def ccx(self, control_qubits: list, target_qubits: int):
        return self.add_triple_gate(control_qubits+[target_qubits], 'ccx')
    
    def show_quantum_circuit(self, plot_labels=True, **kwargs):
        """Use Matplotlib to plot a quantum circuit.
        kwargs    Can override plot_parameters
        """
        labels = []
        inits = {}
        for i in range(self.width):
            labels.append(f"q_{i}")
            inits[f"q_{i}"] = i
        plot_params = dict(scale = 1.0, fontsize = 14.5, linewidth = 2.0, 
                            control_radius = 0.05, not_radius = 0.15, 
                            swap_delta = 0.08, label_buffer = 0.8, 
                            rectangle_delta = 0.3, box_pad=0.2)
        plot_params.update(kwargs)
        scale = plot_params['scale']
        
        # Create labels from gates. This will become slow if there are a lot 
        #  of gates, in which case move to an ordered dictionary
        if not labels:
            labels = []
            for i,gate in tools.enumerate_gates(self.circuit_diag, schedule=True):
                for label in gate[1:]:
                    if label not in labels:
                        labels.append(label)
        
        nq = len(labels)
        nt = len(self.circuit_diag)
        wire_grid = np.arange(0.0, nq*scale, scale, dtype=float)
        gate_grid = np.arange(0.0, nt*scale, scale, dtype=float)
        gate_grid_index = [0.0 for _ in range(nq)]
        # print(gate_grid_index)
        
        fig,ax = tools.setup_figure(nq,nt,gate_grid,wire_grid,plot_params)

        measured = tools.measured_wires(self.circuit_diag, labels, schedule=True)
        tools.draw_wires(ax, nq, gate_grid, wire_grid, plot_params, 'k', measured)
        
        if plot_labels: 
            tools.draw_labels(ax, labels, inits, gate_grid, wire_grid, plot_params, 'k')

        tools.draw_gates(ax, self.circuit_diag, labels, gate_grid_index, wire_grid, plot_params, measured, schedule=True)
        plt.show()
        # return ax
    
    def measure(self, measure_bits_list: Union[list, int]=None, shots: int=1000)->dict:
        """
        测量们
        Args:
            measure_bits_list: 测量比特列表，传入比特位置或列表位置
            shots: 测量次数
        Returns:
            返回测量结果
        """
        # 首先通过执行操作得到所有的状态
        # print(self.gate_list)
        self.run_statevector()
        print(self.state)
        state = np.array(self.state)
        prob = np.real(state.conjugate() * state)
        count = {format(i, f'0{self.width}b'): prob[i] for i in range(len(prob))}
        print('old count {}'.format(count))
        distribution = {}

        if isinstance(measure_bits_list, int):
            measure_bits_list = [measure_bits_list]
        if measure_bits_list is None: # 若为空默认测量所有可能
            print('empty measure list')
            # distribution = count
            measure_bits_list = sorted([_ for _ in range(self.width)], reverse=True)
        
        print('old measure bit {}'.format(measure_bits_list))
        # 计算测量分布
        for p in count.keys():
            # key = ''.join(p[pos - 1] for pos in measure_bits_list)
            key = ''.join(p[pos] for pos in measure_bits_list)
            if count[p] == 0:
                continue
            if key not in distribution:
                distribution[key] = count[p]
            else:
                distribution[key] += count[p]
        print('old distribution {}'.format(distribution))
        # 根据分布抽样概率
        result = {}
        cumulate = 0
        sample = np.random.uniform(0,1, size=shots)
        for key in distribution.keys():
            new_cumulate = cumulate + distribution[key]
            result[key] = sum((cumulate<=sample) & (sample<new_cumulate))
            cumulate = new_cumulate
        return result



# 下面这段代码仅限测试使用，不会参与实际工程
if __name__ == '__main__':
    nQubits = 4
    qc = QuantumCircuit()
    qc.add_qubits(nQubits, name='qft')
    # for i in range(nQubits):
    #     if nQubits - 2 == i:
    #         qc.x(i)
    #     else:
    #         qc.h(i)
    # for i in range(nQubits - 1, -1, -1):
    #     for j in range(i, -1, -1):
    #         if j == i:
    #             qc.h(j)
    #         else:
    #             qc.cp(control_qubit=j, target_qubit=i, theta=np.pi / (2 ** (i - j)))
    qc.h(0)
    # qc.x(1)
    # qc.h(3)
    print(qc.gate_list)
    # for i in range(0, nQubits // 2):
    #     qc.swap(qubit_1=i, qubit_2=nQubits - 1 - i)
    # qc.run_statevector()
    # print(qc.state)
    # print(len(qc.state))
    # measure_pos = sorted([1, 0,2,3], reverse=True)
    # print(measure_pos)
    print(qc.measure())
    # print(qc.measure(measure_bits_list=measure_pos))
    # qc.show_quantum_circuit()    





import torch
import math
import torch_npu
# from tgqSim import QuantumCircuit

# def ActOn_State_PyTorch(psi_real, psi_imag, num_qubits, Gate_type, Gate_pos, *Angles):
#     if torch.npu.is_available():
#         device = torch.device("npu")
#     else:
#         device = torch.device("cpu")

def ActOn_State(psi: list, num_qubits, Gate_type, Gate_pos):
    """
    Apply quantum gates on the state vector using PyTorch.

    :param psi_real: Real part of the state vector
    :param psi_imag: Imaginary part of the state vector
    :param num_qubits: Number of qubits
    :param Gate_type: Type of the gate ('ccx' or 'cswap')
    :param Gate_pos: Positions of the qubits the gate acts on
    :return: Updated real and imaginary parts of the state vector
    """
    psi_real = torch.tensor(psi, dtype=torch.float32)
    psi_imag = torch.tensor([0]*len(psi), dtype=torch.float32)
    device = torch.device("npu" if torch.npu.is_available() else "cpu")
    psi_real = psi_real.to(device)
    psi_imag = psi_imag.to(device)
    
    j2, j1, j0 = sorted(Gate_pos, reverse=True)
    i1_plus = 2 ** j0
    i2_plus = 2 ** j1
    i3_plus = 2 ** j2
    i4_plus = 2 ** j0 + 2 ** j1
    i5_plus = 2 ** j0 + 2 ** j2
    i6_plus = 2 ** j1 + 2 ** j2
    i7_plus = 2 ** j0 + 2 ** j1 + 2 ** j2
    delta3 = 2 ** (j2 + 1)
    delta2 = 2 ** (j1 + 1)
    delta1 = 2 ** (j0 + 1)
    max3 = 2 ** num_qubits - delta3
    max2 = 2 ** j2 - delta2
    max1 = 2 ** j1 - delta1
    max0 = 2 ** j0 - 1

    for k in range(0, max3 + 1, delta3):
        for l in range(0, max2 + 1, delta2):
            for m in range(0, max1 + 1, delta1):
                for n in range(0, max0 + 1):
                    i0 = n | m | l | k
                    i1 = i0 | i1_plus
                    i2 = i0 | i2_plus
                    i3 = i0 | i3_plus
                    i4 = i0 | i4_plus
                    i5 = i0 | i5_plus
                    i6 = i0 | i6_plus
                    i7 = i0 | i7_plus
                    
                    if len(set([i0, i1, i2, i3, i4, i5, i6, i7])) == 8:  # All indices are different
                        if Gate_type == 'ccx':
                            # CCX (Toffoli) gate: flip the target qubit if both control qubits are |1>
                            psi_real[i7], psi_real[i3] = psi_real[i3].clone(), psi_real[i7].clone()
                            psi_imag[i7], psi_imag[i3] = psi_imag[i3].clone(), psi_imag[i7].clone()
                        elif Gate_type == 'cswap':
                            # CSWAP gate: swap the two target qubits if the control qubit is |1>
                            psi_real[i5], psi_real[i6] = psi_real[i6].clone(), psi_real[i5].clone()
                            psi_imag[i5], psi_imag[i6] = psi_imag[i6].clone(), psi_imag[i5].clone()
                            psi_real[i7], psi_real[i3] = psi_real[i3].clone(), psi_real[i7].clone()
                            psi_imag[i7], psi_imag[i3] = psi_imag[i3].clone(), psi_imag[i7].clone()

    return psi_real + 1j*psi_imag



if __name__ == '__main__':
    psi_real = torch.tensor([1, 0, 0, 0, 0, 0, 0, 0], dtype=torch.float32)
    psi_imag = torch.tensor([0, 0, 0, 0, 0, 0, 0, 0], dtype=torch.float32)
    num_qubits = 3
    circuit = QuantumCircuit(num_qubits)

    circuit.add_triple_gate([0, 1, 2], 'ccx')
    circuit.add_triple_gate([0, 1, 2], 'cswap')
    # print(circuit.gate_list)
    circuit.run_statevector()
    print(circuit.state)
    
    gate_list = [
        ([0, 1, 2], ('ccx',)),  # Toffoli gate
        ([0, 1, 2], ('cswap',))  # Controlled-SWAP gate
    ]
    for ele in gate_list:
        Gate_type = ele[1][0]
        Gate_pos = ele[0]
        Angles = tuple()
        if len(ele[1]) > 1:
            Angles = tuple(ele[1][1:])
        psi_real, psi_imag = ActOn_State(psi_real, psi_imag, num_qubits, Gate_type, Gate_pos, *Angles)
    final_psi = psi_real + 1j * psi_imag
    print("final_psi:", final_psi)


import logging
from subprocess import Popen, PIPE
import numpy as np
import os
import ctypes, glob
from tgqSim.utils.env import IS_WINDOWS, IS_LINUX, IS_DARWIN, check_env_flag, check_negative_env_flag
from tgqSim.utils.cuda import find_nvcc, find_cuda_version


# def get_cuda_version():
#     try:
#         # Run nvcc command to get CUDA version
#         p = Popen(["nvcc", "--version"], stdout=PIPE)
#         stdout, _ = p.communicate()
#         # Extract CUDA version from the output
#         output = stdout.decode('utf-8')
#         output_lines = output.split("\n")
#         for line in output_lines:
#             if line.strip().startswith("Cuda compilation tools"):
#                 cuda_version = line.split()[4].rstrip(",")
#                 return cuda_version
#         return None
#     except Exception as e:
#         print("Error:", e)
#         return None

LINUX_HOME = '/usr/local/cuda'
WINDOWS_HOME = glob.glob('C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v*.*')

def get_cuda_version()->str:
    if check_negative_env_flag('USE_CUDA') or check_env_flag('USE_ROCM'):
        # USE_CUDA = False
        CUDA_HOME = None
        CUDA_VERSION = None
    else:
        if IS_LINUX or IS_DARWIN:
            CUDA_HOME = os.getenv('CUDA_HOME', LINUX_HOME)
        else:
            CUDA_HOME = os.getenv('CUDA_PATH', '').replace('\\', '/')
            if CUDA_HOME == '' and len(WINDOWS_HOME) > 0:
                CUDA_HOME = WINDOWS_HOME[0].replace('\\', '/')
        if not os.path.exists(CUDA_HOME):
            # We use nvcc path on Linux and cudart path on macOS
            if IS_LINUX or IS_WINDOWS:
                cuda_path = find_nvcc()
            else:
                cudart_path = ctypes.util.find_library('cudart')
                if cudart_path is not None:
                    cuda_path = os.path.dirname(cudart_path)
                else:
                    cuda_path = None
            if cuda_path is not None:
                CUDA_HOME = os.path.dirname(cuda_path)
            else:
                CUDA_HOME = None
        CUDA_VERSION = find_cuda_version(CUDA_HOME)
        # USE_CUDA = CUDA_HOME is not None
    return ".".join(CUDA_VERSION.split(".")[:2])

def get_computer_cap()->str:
    """
    获取GPU卡计算能力值

    Returns:
        str: 计算能力的数值
    """
    # Run nvidia-smi command to get computer-cap
    p = Popen(["nvidia-smi",  "--query-gpu=compute_cap", "--format=csv,noheader"], stdout=PIPE)
    stdout, _ = p.communicate()
    # Extract CUDA version from the output
    computer_cap = stdout.decode('utf-8').rstrip('\n').replace('.', '')
    return computer_cap

# todo:
def get_dcu_version():
    try:
        p = Popen(["hipcc", "--version"], stdout=PIPE)
        stdout, _ = p.communicate()
        output = stdout.decode('utf-8')
        output_lines = output.split("\n")
        for line in output_lines:
            if line.strip().startswith("HIP version"):
                dcu_version = '.'.join(line.split()[-1].split('.')[:2])
                return dcu_version
        return None
    except Exception as e:
        return None


def get_computer_cap()->str:
    """
    获取GPU卡计算能力值

    Returns:
        str: 计算能力的数值
    """
    # Run nvidia-smi command to get computer-cap
    p = Popen(["nvidia-smi",  "--query-gpu=compute_cap", "--format=csv,noheader"], stdout=PIPE)
    stdout, _ = p.communicate()
    # Extract CUDA version from the output
    computer_cap = stdout.decode('utf-8').rstrip('\n').replace('.', '')
    return computer_cap


def get_normalization(frequency: dict)->dict:
    sum_freq = sum(frequency.values())
    prob = {}
    for key in frequency.keys():
        prob[key] = frequency[key] / sum_freq
    return prob


def free_state(state):
    lib = get_cuda_lib()
    lib.freeAllMem.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.complex128)
    ]
    lib.freeAllMem.restype = None
    lib.freeAllMem(state)


def get_cuda_lib():
    cuda_version = get_cuda_version().replace(".", "-")
    lib_name = f"cuda_{cuda_version}_tgq_simulator.so"
    # computer_cap = get_computer_cap()
    # lib_name = f"cuda_{cuda_version}_sm{computer_cap}_tgq_simulator.so"
    current_file_path = os.path.abspath(__file__)
    current_directory = os.path.dirname(current_file_path)
    dll_path = os.path.abspath(current_directory + '/lib/' + lib_name)
    lib = ctypes.CDLL(dll_path)
    return lib


def get_dcu_lib():
    dcu_version = get_dcu_version().replace(".", "-")
    lib_name = "dcu_{}_tgq_simulator.so".format(dcu_version,)
    current_file_path = os.path.abspath(__file__)
    current_directory = os.path.dirname(current_file_path)
    dll_path = os.path.abspath(current_directory + '/lib/' + lib_name)
    lib = ctypes.CDLL(dll_path)
    return lib


def logger():
    lg = logging.getLogger()
    lg.setLevel(logging.INFO)
    return lg




import os
import glob
import re
from subprocess import Popen, PIPE

from tgqSim.utils.env import IS_WINDOWS

LINUX_HOME = '/usr/local/cuda'
WINDOWS_HOME = glob.glob('C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v*.*')


def find_nvcc():
    if IS_WINDOWS:
        proc = Popen(['where', 'nvcc.exe'], stdout=PIPE, stderr=PIPE)
    else:
        proc = Popen(['which', 'nvcc'], stdout=PIPE, stderr=PIPE)
    out, err = proc.communicate()
    out = out.decode().strip()
    if len(out) > 0:
        if IS_WINDOWS:
            if out.find('\r\n') != -1:
                out = out.split('\r\n')[0]
            out = os.path.abspath(os.path.join(os.path.dirname(out), ".."))
            out = out.replace('\\', '/')
            out = str(out)
        return os.path.dirname(out)
    else:
        return None


def find_cuda_version(cuda_home):
    if cuda_home is None:
        return None
    if IS_WINDOWS:
        candidate_names = [os.path.basename(cuda_home)]
    else:
        # get CUDA lib folder
        cuda_lib_dirs = ['lib64', 'lib']
        for lib_dir in cuda_lib_dirs:
            cuda_lib_path = os.path.join(cuda_home, lib_dir)
            if os.path.exists(cuda_lib_path):
                break
        # get a list of candidates for the version number
        # which are files containing cudart
        candidate_names = list(glob.glob(os.path.join(cuda_lib_path, '*cudart*')))
        candidate_names = [os.path.basename(c) for c in candidate_names]

    # suppose version is MAJOR.MINOR.PATCH, all numbers
    version_regex = re.compile(r'[0-9]+\.[0-9]+\.[0-9]+')
    candidates = [c.group() for c in map(version_regex.search, candidate_names) if c]
    if len(candidates) > 0:
        # normally only one will be retrieved, take the first result
        return candidates[0]
    # if no candidates were found, try MAJOR.MINOR
    version_regex = re.compile(r'[0-9]+\.[0-9]+')
    candidates = [c.group() for c in map(version_regex.search, candidate_names) if c]
    if len(candidates) > 0:
        return candidates[0]


import re
from collections import namedtuple

Token = namedtuple('Token', ['type', 'value'])

def tokenize(code):
    keywords = {'OPENQASM', 'include', 'qreg', 'creg', 'measure', 'gate', 'barrier', 'if', 'h', 'cx'}
    token_specification = [
        ('COMMENT',  r'//.*'),            # Single-line comment
        ('NUMBER',   r'\d+(\.\d*)?'),     # Integer or decimal number
        ('ID',       r'[A-Za-z_]\w*'),    # Identifiers
        ('OP',       r'[{}[\]();,->]'),   # Operators and punctuation
        ('SKIP',     r'[ \t\n]+'),        # Skip spaces and tabs
        ('MISMATCH', r'.'),               # Any other character
    ]
    tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_specification)
    get_token = re.compile(tok_regex).match
    pos = 0
    match = get_token(code)
    while match is not None:
        type_ = match.lastgroup
        if type_ != 'SKIP' and type_ != 'COMMENT':
            val = match.group(type_)
            if type_ == 'ID' and val in keywords:
                type_ = val.upper()
            yield Token(type_, val)
        pos = match.end()
        match = get_token(code, pos)
    if pos != len(code):
        raise RuntimeError('Unexpected character %r on line %d' % (code[pos], code.count('\n', 0, pos) + 1))

class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.pos = 0

    def _consume(self, expected_type=None):
        token = self.tokens[self.pos]
        if expected_type and token.type != expected_type:
            raise SyntaxError(f"Expected {expected_type} but got {token.type}")
        self.pos += 1
        return token

    def parse(self):
        ast = []
        while self.pos < len(self.tokens):
            token = self.tokens[self.pos]
            if token.type == 'OPENQASM':
                self._consume('OPENQASM')
                version = self._consume('NUMBER').value
                ast.append(('OPENQASM', version))
                self._consume('OP')  # Consume the ';'
            elif token.type == 'QREG':
                self._consume('QREG')
                name = self._consume('ID').value
                self._consume('OP')  # Consume the '['
                size = int(self._consume('NUMBER').value)
                self._consume('OP')  # Consume the ']'
                self._consume('OP')  # Consume the ';'
                ast.append(('QREG', name, size))
            elif token.type == 'CREG':
                self._consume('CREG')
                name = self._consume('ID').value
                self._consume('OP')  # Consume the '['
                size = int(self._consume('NUMBER').value)
                self._consume('OP')  # Consume the ']'
                self._consume('OP')  # Consume the ';'
                ast.append(('CREG', name, size))
            elif token.type in {'H', 'CX'}:
                gate = token.type
                self._consume(gate)
                qubits = []
                while self.pos < len(self.tokens) and (self.tokens[self.pos].type != 'OP' or self.tokens[self.pos].value != ';'):
                    if self.tokens[self.pos].type == 'ID' and self.pos + 1 < len(self.tokens) and self.tokens[self.pos + 1].value == '[':
                        qubit_name = self._consume('ID').value
                        self._consume('OP')  # Consume the '['
                        index = int(self._consume('NUMBER').value)
                        self._consume('OP')  # Consume the ']'
                        qubits.append((qubit_name, index))
                    self.pos += 1
                if self.pos < len(self.tokens):
                    self._consume('OP')  # Consume the ';'
                ast.append((gate, qubits))
            elif token.type == 'MEASURE':
                self._consume('MEASURE')
                qubit_name = self._consume('ID').value
                self._consume('OP')  # Consume the '['
                qubit_index = int(self._consume('NUMBER').value)
                self._consume('OP')  # Consume the ']'
                self._consume('OP')  # Consume the '->'
                creg_name = self._consume('ID').value
                self._consume('OP')  # Consume the '['
                creg_index = int(self._consume('NUMBER').value)
                self._consume('OP')  # Consume the ']'
                self._consume('OP')  # Consume the ';'
                ast.append(('MEASURE', (qubit_name, qubit_index), (creg_name, creg_index)))
            else:
                self.pos += 1  # Skip any unexpected tokens
        return ast


code = """
OPENQASM 2.0;

h q[0];
cx q[0], q[1];
measure q[0] ;
measure q[1] ;
"""
tokens = list(tokenize(code))
for token in tokens:
    print(token)

parser = Parser(tokens)
ast = parser.parse()
print(ast)



import torch
import torch_npu
import math
# from tgqSim import QuantumCircuit
# Define complex multiplication function
def complex_mul(a_real, a_imag, b_real, b_imag):
    return [a_real * b_real - a_imag * b_imag, a_real * b_imag + a_imag * b_real]

# Define complex exponential function
def complex_exp(real, imag):
    exp_real = torch.exp(real)
    return [exp_real * torch.cos(imag), exp_real * torch.sin(imag)]

# Function to apply a gate on a quantum state
def ActOn_State(psi, num_qubits, Gate_type, Gate_pos: int, *Angles):
    psi_real = torch.tensor(psi, dtype=torch.float32)
    psi_imag = torch.tensor([0]*len(psi), dtype=torch.float32)
    if torch.npu.is_available():
        device = torch.device("npu")
    else:
        device = torch.device("cpu")
    psi_real = psi_real.to(device)
    psi_imag = psi_imag.to(device)

    # Calculate position indices
    delta = 2 ** (Gate_pos + 1)
    max_delta = 2 ** num_qubits - delta
    i = 1 << Gate_pos

    # Iterate over state vector
    for k in range(0, max_delta + 1, delta):
        for l in range(0, i):
            i0 = l | k
            i1 = i0 | i
            
            # Handle different gate types
            if Gate_type == 'x':
                # Pauli-X (NOT) gate
                psi_real[i0], psi_real[i1] = psi_real[i1].clone(), psi_real[i0].clone()
                psi_imag[i0], psi_imag[i1] = psi_imag[i1].clone(), psi_imag[i0].clone()
            elif Gate_type == 'y':
                # Pauli-Y gate
                psi_real[i0], psi_imag[i0], psi_real[i1], psi_imag[i1] = -psi_imag[i1].clone(), psi_real[i1].clone(), psi_imag[i0].clone(), -psi_real[i0].clone()
            elif Gate_type == 'z':
                # Pauli-Z gate
                psi_real[i1], psi_imag[i1] = -psi_real[i1].clone(), -psi_imag[i1].clone()
            elif Gate_type == 'h':
                # Hadamard gate
                inv_sqrt2 = 1 / math.sqrt(2)
                temp_real_i0 = (psi_real[i0] + psi_real[i1]) * inv_sqrt2
                temp_imag_i0 = (psi_imag[i0] + psi_imag[i1]) * inv_sqrt2
                temp_real_i1 = (psi_real[i0] - psi_real[i1]) * inv_sqrt2
                temp_imag_i1 = (psi_imag[i0] - psi_imag[i1]) * inv_sqrt2
                psi_real[i0], psi_imag[i0], psi_real[i1], psi_imag[i1] = temp_real_i0, temp_imag_i0, temp_real_i1, temp_imag_i1
            elif Gate_type == 'rx':
                # RX gate
                angle = torch.tensor(Angles[0])
                cos_val = torch.cos(angle / 2)
                sin_val = torch.sin(angle / 2)
                temp_real_i0 = cos_val * psi_real[i0] + sin_val * psi_imag[i1]
                temp_imag_i0 = cos_val * psi_imag[i0] - sin_val * psi_real[i1]
                temp_real_i1 = cos_val * psi_real[i1] + sin_val * psi_imag[i0]
                temp_imag_i1 = cos_val * psi_imag[i1] - sin_val * psi_real[i0]
                psi_real[i0], psi_imag[i0], psi_real[i1], psi_imag[i1] = temp_real_i0, temp_imag_i0, temp_real_i1, temp_imag_i1
            elif Gate_type == 'ry':
                # RY gate
                angle = torch.tensor(Angles[0])
                cos_val = torch.cos(angle / 2)
                sin_val = torch.sin(angle / 2)
                temp_real_i0 = cos_val * psi_real[i0] - sin_val * psi_real[i1]
                temp_imag_i0 = cos_val * psi_imag[i0] - sin_val * psi_imag[i1]
                temp_real_i1 = cos_val * psi_real[i1] + sin_val * psi_real[i0]
                temp_imag_i1 = cos_val * psi_imag[i1] + sin_val * psi_imag[i0]
                psi_real[i0], psi_imag[i0], psi_real[i1], psi_imag[i1] = temp_real_i0, temp_imag_i0, temp_real_i1, temp_imag_i1
            elif Gate_type == 'rz':
                # RZ gate
                angle = torch.tensor(Angles[0])
                exp_i1 = complex_exp(torch.tensor(0.), angle / 2)
                exp_i0 = complex_exp(torch.tensor(0.), -angle / 2)
                psi_real[i0], psi_imag[i0] = complex_mul(psi_real[i0], psi_imag[i0], exp_i0[0], exp_i0[1])
                psi_real[i1], psi_imag[i1] = complex_mul(psi_real[i1], psi_imag[i1], exp_i1[0], exp_i1[1])

            elif Gate_type == 'u3':

                theta, phi, lamb = Angles[0], Angles[1], Angles[2]
                theta = torch.tensor(theta, dtype=torch.float32)
                phi = torch.tensor(phi, dtype=torch.float32)
                lamb = torch.tensor(lamb, dtype=torch.float32)

                c = torch.cos(theta / 2)
                s = torch.sin(theta / 2)
                psi_real_i0 = c * psi_real[i0] - s * torch.cos(lamb) * psi_real[i1] + s * torch.sin(lamb) * psi_imag[i1]
                psi_imag_i0 = c * psi_imag[i0] - s * torch.cos(lamb) * psi_imag[i1] - s * torch.sin(lamb) * psi_real[i1]
                psi_real_i1 = s * torch.cos(phi) * psi_real[i0] - s * torch.sin(phi) * psi_imag[i0] + \
                    c * torch.cos(phi + lamb) * psi_real[i1] - c * torch.sin(phi + lamb) * psi_imag[i1]
                psi_imag_i1 = s * torch.cos(phi) * psi_imag[i0] + s * torch.sin(phi) * psi_real[i0] + \
                    c * torch.cos(phi + lamb) * psi_imag[i1] + c * torch.sin(phi + lamb) * psi_real[i1]
                psi_real[i0], psi_imag[i0], psi_real[i1], psi_imag[i1] = psi_real_i0, psi_imag_i0, psi_real_i1, psi_imag_i1
            elif Gate_type == 's':
                psi_real[i1], psi_imag[i1] = -1.0 * psi_imag[i1].clone(), psi_real[i1].clone()
            elif Gate_type == 'sdg':
                psi_imag[i1], psi_real[i1] = -1.0 * psi_real[i1].clone(), psi_imag[i1].clone()
            elif Gate_type == 't':
                angle = torch.tensor(math.pi / 4)
                c = torch.cos(angle)
                s = torch.sin(angle)
                psi_real_i1 = c * psi_real[i1] - s * psi_imag[i1]
                psi_imag_i1 = s * psi_real[i1] + c * psi_imag[i1]
                psi_real[i1], psi_imag[i1] = psi_real_i1, psi_imag_i1
            elif Gate_type == 'tdg':
                angle = torch.tensor(-math.pi / 4)
                c = torch.cos(angle)
                s = torch.sin(angle)
                psi_real_i1 = c * psi_real[i1] - s * psi_imag[i1]
                psi_imag_i1 = s * psi_real[i1] + c * psi_imag[i1]
                psi_real[i1], psi_imag[i1] = psi_real_i1, psi_imag_i1

    return psi_real + 1j*psi_imag

if __name__ == '__main__':
    # Initial state vector |000>
    psi_real = torch.tensor([1, 0, 0, 0, 0, 0, 0, 0], dtype=torch.float32)
    psi_imag = torch.tensor([0, 0, 0, 0, 0, 0, 0, 0], dtype=torch.float32)
    num_qubits = 3  # Number of qubits in the system

    circuit = QuantumCircuit(num_qubits)
    # 制备初态
    circuit.add_single_gate(0, 'x')
    circuit.add_single_gate(1, 'x')
    circuit.add_single_gate(2, 'x')
    circuit.add_single_gate(1, 'h')
    circuit.add_single_gate(0, 'rx', math.pi/4)
    circuit.add_single_gate(1, 'ry', math.pi/3)
    circuit.add_single_gate(2, 'rz', math.pi/2)
    circuit.add_single_gate(0, 'u3', math.pi/4, math.pi/3, math.pi/2)
    circuit.add_single_gate(1, 's')
    circuit.add_single_gate(2, 'sdg')
    circuit.add_single_gate(1, 't')
    circuit.add_single_gate(2, 'tdg')
    circuit.run_statevector()
    print(circuit.state)

    gate_list = [
        ([0], 'x'),
        ([1], 'x'),
        ([2], ('x',)),
        ([1], ('h',)),  
        ([0], ('rx', math.pi/4)),  
        ([1], ('ry', math.pi/3)),  
        ([2], ('rz', math.pi/2)),  
        ([0], ('u3', math.pi/4, math.pi/3, math.pi/2)),
        ([1], ('s',)),
        ([2], ('sdg',)),
        ([1], ('t',)),
        ([2], ('tdg',))  
    ]

    # Apply each gate in the list to the state vector
    for ele in gate_list:
        Gate_pos = ele[0]
        Gate_type = ele[1][0]
        Angles = tuple()
        if len(ele[1]) > 1:
            Angles = tuple(ele[1][1:])
        psi_real, psi_imag = ActOn_State(psi_real, psi_imag, num_qubits, Gate_type, Gate_pos, *Angles)

    # Combine the real and imaginary parts to form the complex state vector
    final_psi = psi_real + 1j * psi_imag
    print("final_psi:", final_psi)
